import json
import os
from enum import Enum, unique

import requests


class InvalidResponseStatusError(BaseException):
    """Raised when the response status is not the one expected."""

    pass


@unique
class ResponseStatus(Enum):
    """Response values and their name."""

    OK = 200  # response included
    Created = 201  # response included
    Accepted = 202  # response included
    NoContent = 204  # response NOT included
    BadRequest = 400  # error response included
    Unauthorized = 401  # error response included
    Forbidden = 403  # error response included
    NotFound = 404  # error response included
    MethodNotAllowed = 405  # error response included
    Conflict = 409  # error response included
    UnsupportedMediaType = 415  # error response included
    TooManyRequests = 429  # error response included
    InternalServerError = 500  # error NOT response included


@unique
class DepositStatus(Enum):
    """All possible zenodo deposit status."""

    DRAFT = "draft"
    PUBLISHED = "published"


@unique
class SortOrder(Enum):
    """Sort order of the results obtained in 'list' methods."""

    BEST_MATCH = "bestmatch"
    MOST_RECENT = "mostrecent"


@unique
class UploadType(Enum):
    """Possible upload values and their name."""

    PUBLICATION = "publication"
    POSTER = "poster"
    PRESENTATION = "presentation"
    DATA_SET = "dataset"
    IMAGE = "image"
    VIDEO = "video"
    SOFTWARE = "software"
    LESSON = "lesson"
    PHYSICAL_OBJECT = "physicalobject"
    OTHER = "other"


@unique
class AccessRight(Enum):
    """Possible upload values and their name."""

    OPEN = "open"
    EMBARGOED = "embargoed"
    RESTRICTED = "restricted"
    CLOSED = "closed"


@unique
class PublicationType(Enum):
    """Possible publication values and their name."""

    ANNOTATION_COLLECTION = "annotationcollection"
    BOOK = "book"
    SECTION = "section"
    CONFERENCE_PAPER = "conferencepaper"
    DATA_MANAGEMENT_PLAN = "datamanagementplan"
    ARTICLE = "article"
    PATENT = "patent"
    PRE_PRINT = "preprint"
    DELIVERABLE = "deliverable"
    MILESTONE = "milestone"
    PROPOSAL = "proposal"
    REPORT = "report"
    SOFTWARE_DOCUMENTATION = "softwaredocumentation"
    TAXONOMIC_TREATMENT = "taxonomictreatment"
    TECHNICAL_NOTE = "technicalnote"
    THESIS = "thesis"
    WORKING_PAPER = "workingpaper"
    OTHER = "other"


class ZenodoEntry(object):
    """Base class for all Zenodo API objects.

     The base class all Zenodo API objects inherit from.
     Base url and access_token stored in object to query the API.

    Attributes:
        entry_dict: The dictionary used to construct the object.
        base_url: The URL used to query the API (Save changes).
        params: Holds access token to authenticate to the API.
    """

    @staticmethod
    def _get_attribute(entry_dict, key, required=False):
        """Checks the dict for keys.

        Args:
            entry_dict:
                The dict to retrieve values from.
            key:
                The key to look for.
            required:
                Boolean to indicate weather to raise Error on missing key.

        Returns:
            Value or None.

        Raises:
            AttributeError: When key not found if required.
        """
        if key in entry_dict.keys():
            return entry_dict[key]
        if required:
            raise AttributeError("Key %s not found but required" % key)
        return None

    def __init__(self, entry_dict, base_url, access_token):
        """Inits @ZenodoEntry.

        Args:
            entry_dict:
                Dictionary for init.
            base_url:
                The base URL of the API to use.
            access_token:
                The authentication token to use for querying the API.
        """
        self.entry_dict = entry_dict
        self.base_url = base_url
        self.params = {"access_token": access_token}

    def to_dict(self):
        """Removes sensitive information from the object and gives back its dictionary representation.

        Returns:
            The dictionary ready for submission via API.
        """
        d = self.__dict__

        if "entry_dict" in d.keys():
            d.pop("entry_dict")
        if "params" in d.keys():
            d.pop("params")
        if "base_url" in d.keys():
            d.pop("base_url")

        # remove None
        d = {k: v for k, v in d.items() if v is not None}

        # treat special objects
        if "metadata" in d.keys():
            if isinstance(d["metadata"], ZenodoMetadata):
                d["metadata"] = self.metadata.to_dict()

        if "files" in d.keys():
            if isinstance(d["files"], ZenodoFile):
                d["files"] = [x.to_dict() for x in self.files]

        if "stats" in d.keys():
            if isinstance(d["stats"], ZenodoRecordStats):
                d["stats"] = self.stats.to_dict()

        return d


class ZenodoMetadata(ZenodoEntry):
    """All possible metadata of a @ZenodoDeposit."""

    @classmethod
    def default_values(
        cls,
        title,
        creators,
        description,
        license,
        version,
        related_identifiers,
        references,
    ):
        default_values = {
            "access_right": AccessRight.OPEN.value,
            "access_right_category": None,
            "creators": creators,
            "description": description,
            "doi": None,
            "license": license,
            "prereserve_doi": "true",
            "publication_date": None,
            "related_identifiers": related_identifiers,
            "relations": None,
            "references": references,
            "resource_type": None,
            "title": title,
            "version": version,
            "upload_type": UploadType.SOFTWARE.value,
        }
        return cls(default_values)

    def __init__(self, entry_dict):
        super().__init__(entry_dict, "", "")
        self.access_right = self._get_attribute(entry_dict, "access_right")
        self.access_right_category = self._get_attribute(
            entry_dict, "access_right_category"
        )
        self.creators = self._get_attribute(entry_dict, "creators")
        self.description = self._get_attribute(entry_dict, "description")
        self.doi = self._get_attribute(entry_dict, "doi")
        self.license = self._get_attribute(entry_dict, "license")
        self.prereserve_doi = self._get_attribute(entry_dict, "prereserve_doi")
        self.publication_date = self._get_attribute(entry_dict, "publication_date")
        self.related_identifiers = self._get_attribute(
            entry_dict, "related_identifiers"
        )
        self.relations = self._get_attribute(entry_dict, "relations")
        self.resource_type = self._get_attribute(entry_dict, "resource_type")
        self.title = self._get_attribute(entry_dict, "title")
        self.upload_type = self._get_attribute(entry_dict, "upload_type")
        self.version = self._get_attribute(entry_dict, "version")
        self.references = self._get_attribute(entry_dict, "references")


class IterableList(list):
    """List for accessing objects in the list by a certain attribute or by the index."""

    def __init__(self, id_attr):
        super().__init__()
        self._id_attr = id_attr

    def __contains__(self, attr):
        try:
            getattr(self, attr)
            return True
        except (AttributeError, TypeError):
            return False

    def __getattr__(self, attr):
        for item in self:
            if getattr(item, self._id_attr) == attr:
                return item
        return list.__getattribute__(self, attr)

    def __getitem__(self, index):
        if isinstance(index, int):
            return list.__getitem__(self, index)

        try:
            return getattr(self, index)
        except AttributeError as e:
            raise IndexError("No item found with id %r" % index) from e


class ZenodoFile(ZenodoEntry):
    """Describes a file in a @ZenodoDeposit or a @ZenodoRecord."""

    _id_attribute_ = "filename"

    @classmethod
    def list_items(cls, deposit):
        out_list = IterableList(cls._id_attribute_)
        out_list.extend(cls.iter_items(deposit))
        return out_list

    @classmethod
    def iter_items(cls, deposit):
        return (f for f in deposit._files)

    def __init__(self, entry_dict):
        super().__init__(entry_dict, "", "")
        self.checksum = self._get_attribute(entry_dict, "checksum")
        self.bucket = self._get_attribute(entry_dict, "bucket")
        self.key = self._get_attribute(entry_dict, "key")
        self.filename = self._get_attribute(entry_dict, "filename")
        self.filesize = self._get_attribute(entry_dict, "filesize")
        self.size = self._get_attribute(entry_dict, "size")
        self.id = self._get_attribute(entry_dict, "id")
        self.type = self._get_attribute(entry_dict, "type")
        self.links = self._get_attribute(entry_dict, "links")

    # todo: write tests
    def get_download_link(self):
        return self.links["self"]


class ZenodoDeposit(ZenodoEntry):
    """The Zenodo Deposit class."""

    def __init__(self, entry_dict, base_url, access_token):
        """Inits the class.

        The dictionary usually comes from a API response. Files and metadata keys will have extra classes.

        Args:
            entry_dict:
                dictionary for init.
            base_url:
                The base URL of the API to use.
            access_token:
                The authentication token to use for querying the API.
        """
        super().__init__(entry_dict, base_url, access_token)
        self.conceptdoi = self._get_attribute(entry_dict, "conceptdoi")
        self.conceptrecid = self._get_attribute(entry_dict, "conceptrecid")
        self.created = self._get_attribute(entry_dict, "created")
        self.doi = self._get_attribute(entry_dict, "doi")
        self.doi_url = self._get_attribute(entry_dict, "doi_url")
        self.id = self._get_attribute(entry_dict, "id")
        self.links = self._get_attribute(entry_dict, "links")
        self.modified = self._get_attribute(entry_dict, "modified")
        self.owner = self._get_attribute(entry_dict, "owner")
        self.record_id = self._get_attribute(entry_dict, "record_id")
        self.state = self._get_attribute(entry_dict, "state")
        self.submitted = self._get_attribute(entry_dict, "submitted")
        self.title = self._get_attribute(entry_dict, "title")
        self.version = self._get_attribute(entry_dict, "version")
        self.related_identifiers = self._get_attribute(
            entry_dict, "related_identifiers"
        )
        self.references = self._get_attribute(entry_dict, "references")

        meta_init = self._get_attribute(entry_dict, "metadata")
        self.metadata = (
            ZenodoMetadata(meta_init) if meta_init is not None else meta_init
        )

        files_init = self._get_attribute(entry_dict, "files")
        self.files = files_init

    @property
    def files(self):
        return ZenodoFile.list_items(self)

    @files.setter
    def files(self, files_init):
        if files_init:
            files = []
            for file_entry in files_init:
                if isinstance(file_entry, ZenodoFile):
                    files.append(file_entry)
                else:
                    files.append(ZenodoFile(file_entry))
            self._files = files
        else:
            self._files = []

    # ############# Deposits attributes #############

    def get_files_id_by_name(self, name):
        """Retrieve the id of a file given its name.

        Args:
            name:
                The name of the file. A full path is not allowed since they are not stored in the Zenodo DB.

        Returns:
            The ID of the file if the name is unique.

        Raises:
            AssertionError: If the name is not unique.
        """
        file_id = None
        unique_control = 0
        if self.files:
            for file_entry in self.files:
                if file_entry.filename == name:
                    file_id = file_entry.id
                    unique_control += 1

        if unique_control > 1:
            raise AssertionError(
                "Deposit broken. File names not unique. Please reload deposit and try again!"
            )

        return file_id

    def list_files_name(self):
        """Obtain a list of all filenames in the deposit.

        Returns:
            list of names. Can be empty.
        """
        names = []
        for file in self.files:
            names.append(file.filename)

        return names

    # ############# Deposits operations #############

    def reload(self):
        """Reload the deposit. Resets all changes.

        Returns:
             True if success.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """
        link = self.base_url + "/api/deposit/depositions/%s" % self.id

        r = requests.get(link, params=self.params)

        response_dict = ZenodoAPI.validate_response(r, ResponseStatus.OK)

        # update object according to response
        self.__init__(response_dict, self.base_url, self.params["access_token"])

        return True

    def update(self, zenodo_metadata):
        """Updates the metadata of a deposit.

        Args:
            zenodo_metadata:
                The metadata. Either as dictionary or as @ZenodoMetadata object.

        Returns:
             True if success.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """
        if not isinstance(zenodo_metadata, ZenodoMetadata):
            try:
                zenodo_metadata = ZenodoMetadata(zenodo_metadata)
            except AttributeError:
                print("Please provide a valid dictionary as input!")

        link = self.base_url + "/api/deposit/depositions/%s" % self.id

        headers = {"Content-Type": "application/json"}
        r = requests.put(
            link,
            params=self.params,
            data=json.dumps({"metadata": zenodo_metadata.to_dict()}),
            headers=headers,
        )

        response_dict = ZenodoAPI.validate_response(r, ResponseStatus.OK)

        # update object according to response
        self.__init__(response_dict, self.base_url, self.params["access_token"])

        return True

    def delete(self):
        """Deletes an unpublished deposit.

        Returns:
             True if success.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """

        # todo: check if not already published

        link = self.base_url + "/api/deposit/depositions/%s" % self.id

        r = requests.delete(link, params=self.params)

        if ZenodoAPI.validate_response(r, ResponseStatus.NoContent):
            self.__init__({}, "", "")

        return True

    # ############# Deposits actions #############

    def publish(self):
        """Publishes an unpublished repo.

        Returns:
             True if success.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """

        # todo: check if already published

        link = self.base_url + "/api/deposit/depositions/%s/actions/publish" % self.id
        r = requests.post(link, params=self.params)

        response_dict = ZenodoAPI.validate_response(r, ResponseStatus.Accepted)

        # update object according to response
        self.__init__(response_dict, self.base_url, self.params["access_token"])

        return True

    def new_version(self):
        """Creates a new version of the deposit.

        Returns:
             True if success.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """

        def extract_draft_id(latest_draft_l):
            draft_id, _ = os.path.splitext(os.path.basename(latest_draft_l))
            return draft_id

        link = (
            self.base_url + "/api/deposit/depositions/%s/actions/newversion" % self.id
        )

        r = requests.post(link, params=self.params)

        response_dict = ZenodoAPI.validate_response(r, ResponseStatus.OK)

        # update object according to response
        self.__init__(response_dict, self.base_url, self.params["access_token"])

        # return new version
        try:
            latest_draft_link = self.links["latest_draft"]
            latest_draft_id = extract_draft_id(latest_draft_link)
        except KeyError as e:
            raise KeyError(
                "Could not find link to latest_draft in response! Aborting..."
            ) from e

        return ZenodoAPI(self.base_url, self.params["access_token"]).deposit_get(
            latest_draft_id, status=DepositStatus.DRAFT
        )

    # ############# Deposition files #############

    def get_remote_files(self):
        """Retrieve all files listed in a deposit

        Returns:
             A list of @ZenodoFile objects.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """
        link = self.base_url + "/api/deposit/depositions/%s/files" % self.id

        r = requests.get(link, params=self.params)

        response_dict = ZenodoAPI.validate_response(r, ResponseStatus.OK)

        file_list = []
        for file_dict in response_dict:
            file_list.append(ZenodoFile(file_dict))

        self.files = file_list if file_list else None

        return file_list

    def create_file(self, file):
        """Uploads a new file to the deposit.

        Args:
            file:
                The full path to the file to upload.

        Returns:
            A @ZenodoFile object.
        """
        link = self.base_url + "/api/deposit/depositions/%s/files" % self.id

        data = {"name": os.path.basename(file)}
        open_file = open(file, "rb")
        files = {"file": open_file}

        r = requests.post(link, data=data, files=files, params=self.params)

        open_file.close()

        response_dict = ZenodoAPI.validate_response(r, ResponseStatus.Created)

        self.get_remote_files()

        return ZenodoFile(response_dict)

    def delete_file_by_id(self, file_id):
        """Deletes the file with a certain ID in the deposit.

        Args:
            file_id:
                The id of the file to delete.

        Returns:
             True if success.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """
        link = self.base_url + "/api/deposit/depositions/%s/files/%s" % (
            self.id,
            file_id,
        )

        r = requests.delete(link, params=self.params)

        ZenodoAPI.validate_response(r, ResponseStatus.NoContent)

        self.get_remote_files()

        return True

    def update_file_by_id(self, file_id, new_file):
        """Updates a existent file with the new version.

        Args:
            file_id:
                The id of the file to delete.
            new_file:
                 The new version of the file.

        Returns:
             The new created @ZenodoFile object.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """
        self.delete_file_by_id(file_id)
        return self.create_file(new_file)

    def delete_file(self, file_name):
        """Deletes a file in a deposit given its name

        Args:
            file_name:
                The name of the file to delete.

        Returns:
             True on success.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """
        file_id = self.get_files_id_by_name(file_name)
        if file_id:
            self.delete_file_by_id(file_id)

    def update_file(self, file_name, new_file):
        """Updates a file in a deposit given its name

        Args:
            file_name:
                The name of the file to delete.
            new_file:
                 The new version of the file.

        Returns:
             The updated @ZenodoFile object.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """
        file_id = self.get_files_id_by_name(file_name)
        if file_id:
            self.update_file_by_id(file_id, new_file)


class ZenodoRecord(ZenodoDeposit):
    """Class for the @ZenodoRecord. Holds the statistics (@ZenodoRecordStats) of a published deposit."""

    def __init__(self, entry_dict, base_url, access_token):
        super().__init__(entry_dict, base_url, access_token)
        self.owners = self._get_attribute(entry_dict, "owners")
        self.revision = self._get_attribute(entry_dict, "revision")
        self.stats = ZenodoRecordStats(self._get_attribute(entry_dict, "stats"))
        self.updated = self._get_attribute(entry_dict, "updated")

    def print_stats(self):
        """Prints the dictionary."""
        print(self.stats.to_dict())


class ZenodoRecordStats(ZenodoEntry):
    """Class holding the statistics of a @ZenodoRecord."""

    def __init__(self, entry_dict):
        super().__init__(entry_dict, "", "")
        self.downloads = self._get_attribute(entry_dict, "downloads")
        self.unique_downloads = self._get_attribute(entry_dict, "unique_downloads")
        self.unique_views = self._get_attribute(entry_dict, "unique_views")
        self.version_downloads = self._get_attribute(entry_dict, "version_downloads")
        self.version_unique_downloads = self._get_attribute(
            entry_dict, "version_unique_downloads"
        )
        self.version_unique_views = self._get_attribute(
            entry_dict, "version_unique_views"
        )
        self.version_views = self._get_attribute(entry_dict, "version_views")
        self.version_volume = self._get_attribute(entry_dict, "version_volume")
        self.views = self._get_attribute(entry_dict, "views")
        self.volume = self._get_attribute(entry_dict, "volume")


class ZenodoDefaultUrl(Enum):
    sandbox_url = "https://sandbox.zenodo.org/"
    url = "https://zenodo.org/"


class ZenodoAPI:
    """
    Zenodo API. Querying the ZenodoAPI.
    """

    def __init__(self, base_url=ZenodoDefaultUrl.url.value, access_token=None):
        """Inits the object given a base_url and an access_token.

        Args:
            base_url:
                base URL for all queries
            access_token:
                The authentication token needed to perform operations.
        """
        self.base_url = base_url
        self.params = {"access_token": ""}

        if access_token:
            self.params["access_token"] = access_token

    @staticmethod
    def validate_response(response, expected_response_code=None):
        """Validates the response from a request.

        Args:
            response:
                The response of a request.
            expected_response_code:
                The expected response specified in the API

        Returns:
            JSON response body or "{'true'}" if no body included in response.

        Raises:
            InvalidResponseStatusError: If query response status other than expected.
        """
        # todo: nice feedback when operation not permitted
        status_code = ResponseStatus(response.status_code)
        if not expected_response_code:
            expected_response_code = status_code

        if status_code in [
            ResponseStatus.OK,
            ResponseStatus.Accepted,
            ResponseStatus.Created,
        ]:
            if status_code != expected_response_code:
                raise InvalidResponseStatusError(
                    "Expected %s got %s"
                    % (expected_response_code.name, status_code.name)
                )
            return response.json()
        elif status_code == ResponseStatus.NoContent:
            if status_code != expected_response_code:
                raise InvalidResponseStatusError(
                    "Expected %s got %s"
                    % (expected_response_code.name, status_code.name)
                )
            return json.dumps(True)
        else:
            print("Error: %s" % status_code.name)
            if status_code != ResponseStatus.InternalServerError:
                print("Detailed message:")
                print(response.json()["message"])
                if response.json()["errors"]:
                    print(response.json()["errors"])

        raise InvalidResponseStatusError(
            "Error '%s' occurred. See Log for detailed information!" % status_code.name
        )

    # ############# Deposits #############

    def deposit_get(
        self,
        deposit_id=None,
        q="",
        status=DepositStatus.PUBLISHED,
        sort=SortOrder.BEST_MATCH,
    ):
        """Retrieve a deposit.

        Args:
            deposit_id:
                The id of the deposit. If None specified other params define a search request.
            q:
                The search query. Holds keywords to search for in the deposit metadata.
            status:
                The deposit status. See @DepositStatus
            sort:
                Sorting of the result. See @SortOrder

        Returns:
            A list of @ZenodoDeposit found. Empty if none found.
        """

        link = self.base_url + "/api/deposit/depositions"

        if deposit_id:
            link = link + "/%s" % deposit_id
            params = self.params
        else:
            params = {
                **{"q": q, "status": status.value, "sort": sort.value},
                **self.params,
            }

        r = requests.get(link, params=params)

        response_dict = self.validate_response(r, ResponseStatus.OK)

        deposit_list = []
        if isinstance(response_dict, list):
            for deposit_dict in response_dict:
                deposit_list.append(
                    ZenodoDeposit(
                        deposit_dict, self.base_url, self.params["access_token"]
                    )
                )
        else:
            deposit_list.append(
                ZenodoDeposit(response_dict, self.base_url, self.params["access_token"])
            )

        return deposit_list

    def deposit_create(self):
        """Creates an empty deposit (already uploaded).

        Returns:
            A ZenodoDeposit object.
        """

        link = self.base_url + "/api/deposit/depositions"

        r = requests.post(link, params=self.params, json={})

        response_dict = self.validate_response(r, ResponseStatus.Created)

        return ZenodoDeposit(response_dict, self.base_url, self.params["access_token"])

    def deposit_create_with_prereserve_doi(self, metadata):
        deposit = self.deposit_create()
        deposit.update(metadata)

        return deposit

    # ############# Records #############

    # todo: write tests
    def records_get(
        self,
        record_id=None,
        q="",
        status=DepositStatus.PUBLISHED,
        sort=SortOrder.BEST_MATCH,
        record_type=UploadType.SOFTWARE,
    ):
        """Retrieve a record or searches through published records.

        Args:
            record_id:
                The id of the record to retrieve. If non specified other parameters define search criteria.
            q:
                The search query. Holds keywords to search for in the deposit metadata.
            status:
                The deposit status. See @DepositStatus.
            sort:
                Sorting of the result. See @SortOrder.
            record_type:
                Record type of the record. See @UploadType.

        Returns:
            Returns the @ZenodoRecord object.
        """

        link = self.base_url + "/api/records"

        if record_id:
            link = link + "/%s" % str(record_id)
            params = self.params
        else:
            params = {
                **{
                    "q": q,
                    "status": status.value,
                    "sort": sort.value,
                    "type": record_type.value,
                },
                **self.params,
            }

        r = requests.get(link, params=params)

        response_dict = self.validate_response(r, ResponseStatus.OK)

        record_list = []
        if isinstance(response_dict, list):
            for record_dict in response_dict:
                record_list.append(
                    ZenodoRecord(
                        record_dict, self.base_url, self.params["access_token"]
                    )
                )
        else:
            record_list.append(
                ZenodoRecord(response_dict, self.base_url, self.params["access_token"])
            )

        return record_list



import operator

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.search_manager import ISearchManager
from album.runner import album_logging

module_logger = album_logging.get_active_logger


class SearchManager(ISearchManager):
    def __init__(self, album: IAlbumController):
        self.album = album

    def search(self, keywords):
        module_logger().debug(
            "Searching with following arguments %s..." % ", ".join(keywords)
        )

        search_index = (
            self.album.collection_manager().get_collection_index().get_all_solutions()
        )
        match_score = {}
        for solution_entry in search_index:
            solution_attrs = solution_entry.setup()
            group, name, version = (
                solution_attrs["group"],
                solution_attrs["name"],
                solution_attrs["version"],
            )
            catalog_id = solution_entry.internal()["catalog_id"]
            catalog_name = self.album.catalogs().get_by_id(catalog_id).name()
            unique_id = ":".join([str(catalog_name), group, name, version])

            # todo: nice searching algorithm here
            for keyword in keywords:
                self._find_matches(keyword, match_score, solution_attrs, unique_id)

        sorted_results = sorted(
            match_score.items(), key=operator.itemgetter(1), reverse=True
        )
        return sorted_results

    def _find_matches(self, keyword, match_score, entry, unique_id):
        if isinstance(entry, str):
            solution_result = keyword in entry
            if solution_result:
                if unique_id in match_score.keys():
                    match_score[unique_id] = match_score[unique_id] + 1
                else:
                    match_score[unique_id] = 1
        if isinstance(entry, dict):
            for name, value in entry.items():
                self._find_matches(keyword, match_score, value, unique_id)
        if isinstance(entry, list):
            for item in entry:
                self._find_matches(keyword, match_score, item, unique_id)



from logging import LogRecord

from album.core.api.model.task import ILogHandler, ITask


class LogHandler(ILogHandler):
    def __init__(self):
        super().__init__()
        self._records = []

    def emit(self, record: LogRecord) -> None:
        self._records.append(record)

    def records(self):
        return self._records


class Task(ITask):
    _id = None
    _method = None
    _args = tuple()
    _log_handler: LogHandler = None
    _status: ITask.Status = None

    def __init__(self, method=None, args=None):
        self._method = method
        if args:
            self._args = args

    def id(self):
        return self._id

    def method(self):
        return self._method

    def args(self):
        return self._args

    def log_handler(self):
        return self._log_handler

    def status(self) -> ITask.Status:
        return self._status

    def set_status(self, status):
        self._status = status

    def set_log_handler(self, handler):
        self._log_handler = handler

    def set_id(self, new_id):
        self._id = new_id



from abc import abstractmethod
from pathlib import Path

from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.api.model.solution import ISolution


class ICollectionSolution:
    @abstractmethod
    def catalog(self) -> ICatalog:
        raise NotImplementedError

    @abstractmethod
    def path(self) -> Path:
        raise NotImplementedError

    @abstractmethod
    def database_entry(self) -> ICollectionIndex.ICollectionSolution:
        raise NotImplementedError

    @abstractmethod
    def coordinates(self) -> ICoordinates:
        raise NotImplementedError

    @abstractmethod
    def loaded_solution(self) -> ISolution:
        raise NotImplementedError

    @abstractmethod
    def set_loaded_solution(self, loaded_solution):
        raise NotImplementedError

    @abstractmethod
    def set_coordinates(self, coordinates):
        raise NotImplementedError

    @abstractmethod
    def set_database_entry(self, database_entry: ICollectionIndex.ICollectionSolution):
        raise NotImplementedError



import logging
from typing import Dict

import colorlog

from album.core.api.model.catalog_updates import ICatalogUpdates
from album.runner.core.api.model.solution import ISolution


def get_solution_as_string(solution: ISolution, solution_path):
    setup = solution.setup()
    res = "Solution details about %s:\n\n" % solution_path
    if setup.title:
        res += "%s\n" % setup.title
        res += "%s\n" % ("=" * len(setup.title))
    if setup.description:
        res += "%s\n\n" % setup.description
    res += "Group            : %s\n" % setup.group
    res += "Name             : %s\n" % setup.name
    res += "Version          : %s" % setup.version
    res += "%s" % get_credit_as_string(solution)
    res += "\n\n"
    if setup.solution_creators:
        res += "Solution creators : %s\n" % ", ".join(setup.solution_creators)
    if setup.license:
        res += "License           : %s\n" % setup.license
    if setup.acknowledgement:
        res += "Acknowledgement   : %s\n" % setup.acknowledgement
    if setup.tags:
        res += "Tags              : %s\n" % ", ".join(setup.tags)
    res += "\n"
    res += "Usage:\n\n"
    res += "  album install %s\n" % solution_path
    res += "  %s\n" % get_solution_run_call_as_string(solution)
    res += "  album test %s\n" % solution.coordinates()
    res += "  album uninstall %s\n" % solution.coordinates()
    res += "\n"
    if setup.args:
        res += "Run parameters:\n\n"
        for arg in setup.args:
            res += "  --%s: %s\n" % (arg["name"], arg["description"])
    return res


def get_solution_run_call_as_string(solution):
    param_example_str = ""
    if solution.setup().args:
        for arg in solution.setup().args:
            if "required" in arg:
                if arg["required"]:
                    param_example_str += "--%s PARAMETER_VALUE " % arg["name"]
    run_call = "album run %s %s" % (solution.coordinates(), param_example_str)
    return run_call


def get_credit_as_string(solution: ISolution) -> str:
    res = ""
    if solution.setup().cite:
        res += "\n\nCredits:\n\n"
        for citation in solution.setup().cite:
            text = get_citation_as_string(citation)
            res += "%s\n" % text
        res += "\n"
    return res


def get_citation_as_string(citation):
    text = citation["text"]
    if "doi" in citation and "url" in citation:
        text += " (DOI: %s, %s)" % (citation["doi"], citation["url"])
    else:
        if "doi" in citation:
            text += " (DOI: %s)" % citation["doi"]
        if "url" in citation:
            text += " (%s)" % citation["url"]
    return text


def get_updates_as_string(updates: Dict[str, ICatalogUpdates]) -> str:
    res = ""
    for catalog_name in updates:
        change = updates[catalog_name]
        res += "Catalog: %s\n" % change.catalog().name()
        if len(change.catalog_attribute_changes()) > 0:
            res += "  Catalog attribute changes:\n"
            for item in change.catalog_attribute_changes():
                res += "  name: %s, new value: %s\n" % (
                    item.attribute(),
                    item.new_value(),
                )
        if len(change.solution_changes()) > 0:
            res += "  Catalog solution changes:\n"
            for i, item in enumerate(change.solution_changes()):
                if i is len(change.solution_changes()) - 1:
                    res += "  └─ [%s] %s\n" % (
                        item.change_type().name,
                        item.coordinates(),
                    )
                    separator = " "
                else:
                    res += "  ├─ [%s] %s\n" % (
                        item.change_type().name,
                        item.coordinates(),
                    )
                    separator = "|"
                res += "  %s     %schangelog: %s\n" % (
                    separator,
                    (" " * len(item.change_type().name)),
                    item.change_log(),
                )

        if (
            len(change.catalog_attribute_changes()) == 0
            and len(change.solution_changes()) == 0
        ):
            res += "  No changes.\n"
    return res


def get_index_as_string(index_dict: dict):
    res = "\n"
    if "base" in index_dict:
        res += "Album base directory: %s\n" % index_dict["base"]
    res += "Catalogs in your local collection:\n"
    if "catalogs" in index_dict:
        for catalog in index_dict["catalogs"]:
            res += "Catalog '%s':\n" % catalog["name"]
            res += "├─ name: %s\n" % catalog["name"]
            res += "├─ src: %s\n" % catalog["src"]
            res += "├─ catalog_id: %s\n" % catalog["catalog_id"]
            if len(catalog["solutions"]) > 0:
                res += "├─ deletable: %s\n" % catalog["deletable"]
                res += "└─ [installed] solutions:\n"
                for i, solution in enumerate(catalog["solutions"]):
                    installed = " "
                    if solution["internal"]["installed"]:
                        installed = "x"
                    if i is len(catalog["solutions"]) - 1:
                        res += "   └─ [%s] %s:%s:%s\n" % (
                            installed,
                            solution["setup"]["group"],
                            solution["setup"]["name"],
                            solution["setup"]["version"],
                        )
                    else:
                        res += "   ├─ [%s] %s:%s:%s\n" % (
                            installed,
                            solution["setup"]["group"],
                            solution["setup"]["name"],
                            solution["setup"]["version"],
                        )
            else:
                res += "└─ deletable: %s\n" % catalog["deletable"]
    return res


def get_search_result_as_string(args, search_result):
    res = ""
    if len(search_result) > 0:
        res += (
            'Search results for "%s" - run `album info SOLUTION_ID` for more information:\n'
            % " ".join(args.keywords)
        )
        res += "[SCORE] SOLUTION_ID\n"
        for result in search_result:
            res += "[%s] %s\n" % (result[1], result[0])
    else:
        res += 'No search results for "%s".' % " ".join(args.keywords)
    return res


def get_logging_formatter(fmt=None, time=None):
    if not fmt:
        fmt = "%(log_color)s%(asctime)s %(levelname)-7s %(shortened_name)s%(message)s"
    if not time:
        time = "%H:%M:%S"
    return colorlog.ColoredFormatter(
        fmt,
        time,
        log_colors={
            "DEBUG": "cyan",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red,bold",
        },
    )


def get_logger_name_minimizer_filter():
    class NameDotFilter(logging.Filter):
        def filter(self, record):
            count = record.name.count(".") + record.name.count("~")
            if count > 0:
                record.shortened_name = "~" * count + " "
            else:
                record.shortened_name = ""
            return True

    return NameDotFilter()


def get_message_filter():
    class MessageFilter(logging.Filter):
        def filter(self, record):
            self._apply_mamba_menuinst_filter(record)
            return True

        def _apply_mamba_menuinst_filter(self, record):
            if (
                record.msg
                and isinstance(record.msg, str)
                and "menuinst called from non-root env" in record.msg
            ):
                record.levelname = "WARNING"
                record.levelno = logging.getLevelName("WARNING")

    return MessageFilter()



import gc
import sqlite3
import threading
from abc import ABC
from pathlib import Path

from album.core.api.model.database import IDatabase


class Database(IDatabase, ABC):
    """Abstract model class for all album sqlite databases"""

    def __init__(self, path):
        self.connections = {}
        self.cursors = {}

        self.path = Path(path)

        if not self.is_created(close=False):
            self.create()

    def __del__(self):
        self.close()

    def close_current_connection(self, commit=True):
        current_thread_id = threading.current_thread().ident
        if current_thread_id in self.cursors:
            cursor = self.cursors.pop(current_thread_id)
            cursor.close()

        if current_thread_id in self.connections:
            conn = self.connections.pop(current_thread_id)

            if commit:
                conn.commit()

            # finally close
            conn.close()

    def close(self):
        for cursor_id in self.cursors:
            cursor = self.cursors[cursor_id]
            try:
                cursor.close()
            except sqlite3.ProgrammingError:
                pass
            del cursor

        for thread_id in self.connections:
            connection = self.connections[thread_id]
            try:
                connection.close()
            except sqlite3.ProgrammingError:
                pass
            del connection

        gc.collect(2)
        self.cursors = {}
        self.connections = {}

    def get_connection(self):
        thread_id = threading.current_thread().ident
        if thread_id in self.connections:
            return self.connections[thread_id]
        con = self._create_connection()
        self.connections[thread_id] = con
        return con

    def get_cursor(self):
        thread_id = threading.current_thread().ident
        if thread_id in self.cursors:
            return self.cursors[thread_id]
        cursor = self.get_connection().cursor()
        self.cursors[thread_id] = cursor
        return cursor

    def _create_connection(self):
        con = sqlite3.connect(str(self.path))
        con.row_factory = sqlite3.Row
        return con

    def next_id(self, table_name, close=False):
        cursor = self.get_cursor()

        table_name_id = table_name + "_id"
        is_empty = (
            False
            if cursor.execute("SELECT * FROM %s" % table_name).fetchone()
            else True
        )
        if is_empty:
            return 1

        # note: always use subquery for count/max etc. operations as sqlite python API requires full rows back!
        r = cursor.execute(
            "SELECT * FROM %s WHERE %s = (SELECT MAX(%s) FROM %s)"
            % (table_name, table_name_id, table_name_id, table_name)
        ).fetchone()

        if close:
            cursor.connection.close()

        return int(r[table_name_id]) + 1

    def is_created(self, close=True):
        cursor = self.get_cursor()
        r = cursor.execute("SELECT * FROM sqlite_master").fetchall()
        created = False

        for row in r:
            if row["name"] != "sqlite_sequence":
                created = True
                break

        if close:
            cursor.connection.close()

        return created

    def is_table_empty(self, table, close=True):
        cursor = self.get_cursor()

        r = cursor.execute("SELECT * FROM %s" % table).fetchone()

        if close:
            self.close_current_connection()

        return False if r else True

    def get_path(self):
        return self.path



import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional, Dict

import validators

from album.core.api.controller.collection.catalog_handler import ICatalogHandler
from album.core.api.controller.controller import IAlbumController
from album.core.api.model.catalog import ICatalog
from album.core.model.catalog import Catalog, retrieve_index_files_from_src
from album.core.model.catalog_index import CatalogIndex
from album.core.model.catalog_updates import CatalogUpdates, SolutionChange, ChangeType
from album.core.model.collection_index import CollectionIndex
from album.core.model.default_values import DefaultValues
from album.core.utils.operations.dict_operations import str_to_dict
from album.core.utils.operations.file_operations import (
    force_remove,
    copy,
    get_dict_from_json,
)
from album.core.utils.operations.resolve_operations import dict_to_coordinates
from album.runner import album_logging

module_logger = album_logging.get_active_logger


class CatalogHandler(ICatalogHandler):
    """Helper class responsible for catalog handling."""

    def __init__(self, album: IAlbumController):
        self.album = album

    def create_cache_catalog(self):
        name = DefaultValues.cache_catalog_name.value
        local_path = self.album.configuration().get_cache_path_catalog(name)

        # cache catalog always of type direct, deployment not possible anyways
        catalog_meta_information = self.create_new_metadata(local_path, name, "direct")
        catalog = self._create_catalog_from_src(
            local_path, catalog_meta_information, "main", deletable=False
        )
        self._add_to_index(catalog)
        self.album.migration_manager().load_index(catalog)
        self._update_collection_from_catalog(catalog)

    def add_initial_catalogs(self):
        self.create_cache_catalog()
        initial_catalogs = self.album.configuration().get_initial_catalogs()
        initial_catalogs_branch_name = (
            self.album.configuration().get_initial_catalogs_branch_name()
        )
        for catalog in initial_catalogs.keys():
            self.add_by_src(
                str(initial_catalogs[catalog]), initial_catalogs_branch_name[catalog]
            ).dispose()

    def add_by_src(self, source, branch_name="main") -> Catalog:
        # source can be path or url
        source = str(source)
        if not validators.url(source):
            if Path(source).exists():
                source = str(Path(source).resolve())

        catalog_dict = self._get_collection_index().get_catalog_by_src(source)
        if catalog_dict:
            module_logger().warning("Cannot add catalog twice! Doing nothing...")
            return self._as_catalog(catalog_dict)

        catalog_meta_information = self._retrieve_catalog_meta_information(
            source, branch_name
        )

        catalog = self._create_catalog_from_src(
            source, catalog_meta_information, branch_name
        )

        # always keep the local copy up to date
        if not catalog.is_cache():
            self.album.migration_manager().migrate_catalog_index_db(
                catalog.index_file_path(),  # the path to the catalog
                catalog_meta_information[
                    "version"
                ],  # eventually outdated remote version
                CatalogIndex.version,  # current version in the library
            )

        self._add_to_index(catalog)
        self._create_catalog_cache_if_missing(catalog)
        self.album.migration_manager().load_index(catalog)

        module_logger().info("Catching catalog content..")
        self._update(catalog)
        self._update_collection_from_catalog(catalog)
        module_logger().info("Added catalog %s!" % source)

        return catalog

    def _add_to_index(self, catalog: ICatalog) -> int:
        catalog_id = self._get_collection_index().insert_catalog(
            catalog.name(),
            str(catalog.src()),
            str(catalog.path()),
            int(catalog.is_deletable()),
            str(catalog.branch_name()),
            str(catalog.type()),
        )
        catalog.set_catalog_id(catalog_id)
        return catalog.catalog_id()

    def get_by_id(self, catalog_id) -> Catalog:
        catalog = self._get_collection_index().get_catalog(catalog_id)
        if not catalog:
            raise LookupError('Catalog with id "%s" not configured!' % catalog_id)
        return self._as_catalog(catalog)

    def _get_collection_index(self):
        return self.album.collection_manager().get_collection_index()

    def get_by_src(self, src) -> Catalog:
        src = str(src)
        catalog_dict = self._get_collection_index().get_catalog_by_src(src)
        if not catalog_dict:
            raise LookupError('Catalog with src "%s" not configured!' % src)
        return self._as_catalog(catalog_dict)

    def get_by_name(self, name) -> Catalog:
        """Looks up a catalog by its id and returns it."""
        name = str(name)
        catalog_dict = self._get_collection_index().get_catalog_by_name(name)
        if not catalog_dict:
            raise LookupError('Catalog with name "%s" not configured!' % name)
        return self._as_catalog(catalog_dict)

    def get_by_path(self, path) -> Catalog:
        path = str(path)
        catalog_dict = self._get_collection_index().get_catalog_by_path(path)
        if not catalog_dict:
            raise LookupError('Catalog with path "%s" not configured!' % path)
        return self._as_catalog(catalog_dict)

    def get_all(self) -> List[Catalog]:
        catalogs = []
        catalog_list = self._get_collection_index().get_all_catalogs()

        for catalog_entry in catalog_list:
            catalogs.append(self._as_catalog(catalog_entry))

        return catalogs

    def get_cache_catalog(self) -> Catalog:
        local_catalog = None
        for catalog in self.get_all():
            if catalog.is_local:
                local_catalog = catalog
                break

        if local_catalog is None:
            raise RuntimeError(
                "Misconfiguration of catalogs. There must be at least one local catalog!"
            )

        return local_catalog

    def create_new_metadata(self, local_path, name, catalog_type):
        local_path = Path(local_path)
        if not local_path.exists():
            local_path.mkdir(parents=True)

        meta_data = (
            '{"name": "'
            + name
            + '", "version": "'
            + CatalogIndex.version
            + '", "type": "'
            + catalog_type
            + '"}'
        )
        with open(
            local_path.joinpath(DefaultValues.catalog_index_metafile_json.value), "w"
        ) as meta:
            meta.writelines(meta_data)

        return str_to_dict(meta_data)

    def _update(self, catalog: Catalog) -> bool:
        r = self.album.migration_manager().refresh_index(catalog)
        module_logger().info("Updated catalog %s!" % catalog.name())
        return r

    def update_by_name(self, catalog_name) -> bool:
        catalog = self.get_by_name(catalog_name)

        return self._update(catalog)

    def update_all(self) -> List[bool]:
        catalog_r = []
        for catalog in self.get_all():
            try:
                r = self._update(catalog)
                catalog_r.append(r)
            except Exception:
                module_logger().warning("Failed to update catalog %s!" % catalog.name())
                catalog_r.append(False)
                pass

        return catalog_r

    def update_any(self, catalog_name=None):
        if catalog_name:
            self.update_by_name(catalog_name)
        else:
            self.update_all()

    def update_collection(
        self, catalog_name=None, dry_run: bool = False, override: bool = False
    ) -> Dict[str, CatalogUpdates]:
        if dry_run:
            if catalog_name:
                catalog = self.get_by_name(catalog_name)
                return {
                    catalog_name: self._get_divergence_between_catalog_and_collection(
                        catalog
                    )
                }
            else:
                return self._get_divergence_between_catalogs_and_collection()
        else:
            if catalog_name:
                catalog = self.get_by_name(catalog_name)
                return {
                    catalog_name: self._update_collection_from_catalog(
                        catalog, override
                    )
                }
            else:
                return self._update_collection_from_catalogs(override)

    def _remove_from_collection(self, catalog_dict) -> Optional[Catalog]:
        try:
            catalog_to_remove = self.get_by_id(catalog_dict["catalog_id"])
        except LookupError as err:
            raise LookupError("Cannot remove catalog! Not configured...") from err

        if not catalog_to_remove.is_deletable():
            raise AttributeError("Cannot remove catalog! Marked as not deletable!")

        installed_solutions = self.get_installed_solutions(catalog_to_remove)

        if installed_solutions:
            installed_solutions_string = ", ".join(
                [str(dict_to_coordinates(i.setup())) for i in installed_solutions]
            )
            raise RuntimeError(
                "Cannot remove catalog! "
                "Has the following solutions installed: %s" % installed_solutions_string
            )

        self._get_collection_index().remove_catalog(catalog_to_remove.catalog_id())

        force_remove(catalog_to_remove.path())

        catalog_to_remove.dispose()

        return catalog_to_remove

    def remove_from_collection_by_path(self, path) -> Optional[Catalog]:
        catalog_dict = self._get_collection_index().get_catalog_by_path(path)

        if not catalog_dict:
            module_logger().warning(
                "Cannot remove catalog, catalog with path %s not found!" % str(path)
            )
            return None

        catalog_to_remove = self._remove_from_collection(catalog_dict)

        module_logger().info("Removed catalog with path %s." % str(path))

        return catalog_to_remove

    def remove_from_collection_by_name(self, name) -> Optional[Catalog]:
        catalog_dict = self._get_collection_index().get_catalog_by_name(name)

        if not catalog_dict:
            module_logger().warning(
                'Cannot remove catalog, catalog with name "%s", not found!' % str(name)
            )
            return None

        catalog_to_remove = self._remove_from_collection(catalog_dict)

        module_logger().info("Removed catalog with name %s." % str(name))

        return catalog_to_remove

    def remove_from_collection_by_src(self, src) -> Optional[Catalog]:
        if not validators.url(str(src)):
            if Path(src).exists():
                src = str(Path(src).resolve())
            else:
                module_logger().warning(
                    'Cannot remove catalog with source "%s"! Not configured!' % str(src)
                )
                return None

        catalog_dict = self._get_collection_index().get_catalog_by_src(src)

        if not catalog_dict:
            module_logger().warning(
                'Cannot remove catalog with source "%s"! Not configured!' % str(src)
            )
            return None

        catalog_to_remove = self._remove_from_collection(catalog_dict)

        module_logger().info("Removed catalog with source %s" % str(src))

        return catalog_to_remove

    def get_installed_solutions(self, catalog: ICatalog):
        installed_solutions = (
            self._get_collection_index().get_all_installed_solutions_by_catalog(
                catalog.catalog_id()
            )
        )
        return installed_solutions

    def get_all_as_dict(self) -> dict:
        return {"catalogs": self._get_collection_index().get_all_catalogs()}

    def set_version(self, catalog: ICatalog):
        # database version
        database_version = catalog.index().get_version()

        # cache version
        meta_dict = get_dict_from_json(catalog.get_meta_file_path())
        meta_version = meta_dict["version"]

        if database_version != meta_version:
            raise ValueError(
                f"Catalog meta information (version {meta_version}) unequal to actual version {database_version}!"
            )

        catalog.set_version(database_version)
        return database_version

    def _retrieve_catalog_meta_information(self, source, branch_name="main"):
        with TemporaryDirectory(dir=self.album.configuration().tmp_path()) as tmp_dir:
            repo = Path(tmp_dir).joinpath("repo")
            try:
                _, meta_src = retrieve_index_files_from_src(
                    source, branch_name=branch_name, tmp_dir=repo
                )
                meta_file = copy(
                    meta_src,
                    self.album.configuration()
                    .cache_path_download()
                    .joinpath(DefaultValues.catalog_index_metafile_json.value),
                )
            finally:
                force_remove(repo)

        meta_dict = get_dict_from_json(meta_file)

        return meta_dict

    def _create_catalog_from_src(
        self, src, catalog_meta_information, branch_name="main", deletable=True
    ) -> Catalog:
        """Creates the local cache path for a catalog given its src. (Network drive, git-link, etc.)"""
        # the path where the catalog lives based on its metadata
        catalog_path = self.album.configuration().get_cache_path_catalog(
            catalog_meta_information["name"]
        )

        catalog = Catalog(
            None,
            catalog_meta_information["name"],
            catalog_path,
            src=src,
            branch_name=branch_name,
            catalog_type=catalog_meta_information["type"],
            deletable=deletable,
        )

        catalog.dispose()

        return catalog

    @staticmethod
    def _create_catalog_cache_if_missing(catalog):
        """Creates the path of a catalog if it is missing."""
        if not catalog.path().exists():
            catalog.path().mkdir(parents=True)

    def _get_divergence_between_catalogs_and_collection(
        self,
    ) -> Dict[str, CatalogUpdates]:
        """Gets the divergence list between all catalogs and the catalog_collection."""
        res = {}
        for catalog in self.get_all():
            res[catalog.name()] = self._get_divergence_between_catalog_and_collection(
                catalog=catalog
            )
        return res

    def _get_divergence_between_catalog_and_collection(
        self, catalog: Catalog
    ) -> CatalogUpdates:
        """Gets the divergence between a given catalog and the catalog_collection"""

        if catalog.is_cache():
            # cache catalog is always up to date since src and path are the same
            return CatalogUpdates(catalog)

        solutions_in_collection = self._get_collection_index().get_solutions_by_catalog(
            catalog.catalog_id()
        )
        self.album.migration_manager().load_index(catalog)
        solutions_in_catalog = catalog.index().get_all_solutions()
        solution_changes = self._compare_solutions(
            solutions_in_collection, solutions_in_catalog
        )
        return CatalogUpdates(catalog, solution_changes=solution_changes)

    def _update_collection_from_catalogs(
        self, override: bool = False
    ) -> Dict[str, CatalogUpdates]:
        res = {}
        for catalog in self.get_all():
            res[catalog.name()] = self._update_collection_from_catalog(
                catalog, override
            )
        return res

    def _update_collection_from_catalog(
        self, catalog: Catalog, override: bool = False
    ) -> CatalogUpdates:
        divergence = self._get_divergence_between_catalog_and_collection(catalog)
        # TODO apply changes to catalog attributes
        for change in divergence.solution_changes():
            self.album.solutions().apply_change(divergence.catalog(), change, override)
        return divergence

    @staticmethod
    def _compare_solutions(
        solutions_old: List[CollectionIndex.CollectionSolution],
        solutions_new: List[dict],
    ) -> List[SolutionChange]:
        res = []
        # CAUTION: solutions should not be compared based on their id as this might change

        dict_old_coordinates = {}
        list_old_hash = list([s.internal()["hash"] for s in solutions_old])
        for s in solutions_old:
            dict_old_coordinates[dict_to_coordinates(s.setup())] = s

        dict_new_coordinates = {}
        dict_new_hash = {}
        for s in solutions_new:
            dict_new_coordinates[dict_to_coordinates(s)] = s
            dict_new_hash[s["hash"]] = s

        for hash in dict_new_hash.keys():
            # get metadata
            coordinates = dict_to_coordinates(dict_new_hash[hash])

            # solution updated or added
            if hash not in list_old_hash:
                if (
                    coordinates in dict_old_coordinates.keys()
                ):  # changed when metadata in old solution list
                    change = SolutionChange(
                        coordinates,
                        ChangeType.CHANGED,
                        solution_status=dict_old_coordinates[coordinates].internal(),
                    )
                    res.append(change)

                    # remove solution from old coordinates list
                    # for later iteration over this list to determine deleted solutions
                    dict_old_coordinates.pop(coordinates)

                else:  # added when metadata not in old solution list
                    change = SolutionChange(coordinates, ChangeType.ADDED)
                    res.append(change)
            else:  # not changed
                dict_old_coordinates.pop(coordinates)

        # iterate over remaining solutions in dict_old_coordinates. These are the removed solutions
        for coordinates in dict_old_coordinates.keys():
            change = SolutionChange(coordinates, ChangeType.REMOVED)
            res.append(change)

        return res

    @staticmethod
    def _as_catalog(catalog_dict) -> Catalog:
        return Catalog(
            catalog_dict["catalog_id"],
            catalog_dict["name"],
            catalog_dict["path"],
            catalog_dict["src"],
            bool(catalog_dict["deletable"]),
            catalog_dict["branch_name"],
            catalog_dict["type"],
        )



from pathlib import Path

from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.api.model.collection_solution import ICollectionSolution
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.api.model.solution import ISolution


class ResolveResult(ICollectionSolution):
    def __init__(
        self,
        path,
        catalog: ICatalog,
        collection_entry: ICollectionIndex.ICollectionSolution,
        coordinates: ICoordinates,
        loaded_solution=None,
    ):
        self._catalog: ICatalog = catalog
        self._path = path
        self._collection_entry: ICollectionIndex.ICollectionSolution = collection_entry
        self._coordinates: ICoordinates = coordinates
        self._loaded_solution: ISolution = loaded_solution

    def __eq__(self, other):
        return (
            isinstance(other, ICollectionSolution)
            and other.catalog() == self._catalog
            and other.path() == self._path
            and other.loaded_solution() == self._loaded_solution
            and other.database_entry() == self._collection_entry
            and other.coordinates() == self._coordinates
        )

    def catalog(self) -> ICatalog:
        return self._catalog

    def path(self) -> Path:
        return self._path

    def database_entry(self) -> ICollectionIndex.ICollectionSolution:
        return self._collection_entry

    def coordinates(self) -> ICoordinates:
        return self._coordinates

    def loaded_solution(self) -> ISolution:
        return self._loaded_solution

    def set_loaded_solution(self, loaded_solution):
        self._loaded_solution = loaded_solution

    def set_coordinates(self, coordinates):
        self._coordinates = coordinates

    def set_database_entry(self, database_entry: ICollectionIndex.ICollectionSolution):
        self._collection_entry = database_entry



import argparse
from queue import Queue, Empty
from typing import List

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.script_manager import IScriptManager
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.controller.run_manager import SolutionGroup
from album.core.model.script_queue_entry import ScriptQueueEntry
from album.core.utils.operations.resolve_operations import build_resolve_string
from album.core.utils.operations.solution_operations import (
    get_steps_dict,
    get_parent_dict,
)
from album.runner.album_logging import get_active_logger
from album.runner.core.api.model.script_creator import IScriptCreator
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.script_creator import (
    ScriptCreatorRun,
    ScriptCreatorRunWithParent,
)

module_logger = get_active_logger


class ScriptManager(IScriptManager):
    def __init__(self, album: IAlbumController):
        self.album = album

    def run_solution_script(
        self, resolve_result: ICollectionSolution, script_creator: ScriptCreatorRun
    ):
        queue = Queue()
        self.build_queue(resolve_result, queue, script_creator, False, [""])
        script_queue_entry = queue.get(block=False)
        self.album.environment_manager().run_scripts(
            script_queue_entry.environment,
            script_queue_entry.scripts,
            pipe_output=False,
        )

    def run_queue(self, queue: Queue):
        module_logger().debug("Running queue...")
        try:
            while True:
                script_queue_entry = queue.get(block=False)
                module_logger().debug(
                    'Running task "%s"...' % script_queue_entry.coordinates.name()
                )
                self._run_in_environment(script_queue_entry)
                module_logger().debug(
                    'Finished running task "%s"!'
                    % script_queue_entry.coordinates.name()
                )
                queue.task_done()
        except Empty:
            module_logger().debug("Currently nothing more to run!")

    def build_queue(
        self,
        collection_solution: ICollectionSolution,
        queue,
        script_creator: ScriptCreatorRun,
        run_immediately=False,
        argv=None,
    ):
        if argv is None:
            argv = [""]
        solution = collection_solution.loaded_solution()
        steps = get_steps_dict(solution)
        if steps:  # solution consists of at least one step
            # a step base album is first initialized in the album environment to be able to harvest it's arguments
            # active_solution.init() THIS FEATURE IS TEMPORARY DISABLED

            step_solution_parsed_args = self.__parse_args(solution, argv)
            module_logger().debug("Building queue for %s steps.." % len(steps))

            for i, step in enumerate(steps):
                module_logger().debug(
                    "Adding step %s / %s to queue..." % (i, len(steps))
                )
                if type(step) is list:
                    self._build_steps_queue(
                        queue,
                        step,
                        script_creator,
                        run_immediately,
                        step_solution_parsed_args,
                    )
                else:
                    self._build_steps_queue(
                        queue,
                        [step],
                        script_creator,
                        run_immediately,
                        step_solution_parsed_args,
                    )
        else:  # single element queue, no steps
            parent = get_parent_dict(solution)
            if parent:
                module_logger().debug(
                    "Adding standalone solution with parent to queue..."
                )
                # create script with parent
                queue.put(
                    self._create_solution_run_with_parent_script_standalone(
                        collection_solution, argv, script_creator
                    )
                )
            else:
                module_logger().debug("Adding standalone to queue...")
                # create script without parent
                queue.put(
                    self._create_solution_run_script_standalone(
                        collection_solution, argv, script_creator
                    )
                )

    def _build_steps_queue(
        self,
        queue: Queue,
        steps: list,
        script_creator: ScriptCreatorRun,
        run_immediately=False,
        step_solution_parsed_args=None,
    ):
        """Builds the queue of step-album to be executed. FIFO queue expected.

        Args:
            script_creator:
                The ScriptCreatorRun object to use to create the execution script.
            step_solution_parsed_args:
                Namespace object from parsing the step-solution arguments
            queue:
                The queue object.
            steps:
                The steps of a stepped solution.
            run_immediately:
                Boolean. When true, a collection is run immediately without solving for further steps and pushing them
                to the queue. Can result in resolving problems in further downstream collections. Necessary for
                branching decisions.

        """
        # start with an empty collection of steps with the same parent
        same_parent_step_collection = SolutionGroup(step_solution_parsed_args)
        citations = []

        for step in steps:
            module_logger().debug('resolving step "%s"...' % step["name"])
            resolve_result = self.album.collection_manager().resolve_installed_and_load(
                build_resolve_string(step)
            )

            citations.append(resolve_result.loaded_solution())

            parent = get_parent_dict(resolve_result.loaded_solution())
            if parent:  # collect steps as long as they have the same parent
                current_parent_script_resolve = (
                    self.album.collection_manager().resolve_installed_and_load(
                        build_resolve_string(parent)
                    )
                )

                # check whether the step has the same parent as the previous steps
                if (
                    same_parent_step_collection.parent
                    and same_parent_step_collection.parent.coordinates()
                    != current_parent_script_resolve.coordinates()
                ):

                    # put old collection to queue
                    queue.put(
                        self._create_solution_run_collection_script(
                            same_parent_step_collection, script_creator
                        )
                    )

                    # runs the old collection immediately
                    if run_immediately:
                        self.run_queue(queue)

                    # set new parent
                    same_parent_step_collection.parent = current_parent_script_resolve

                    # overwrite old steps
                    same_parent_step_collection.steps_solution = [
                        resolve_result.loaded_solution()
                    ]
                    same_parent_step_collection.steps = [step]

                else:  # same or new parent
                    module_logger().debug('Pushed step "%s" in queue...' % step["name"])

                    # set parent
                    same_parent_step_collection.parent = current_parent_script_resolve

                    # append another step to the steps already having the same parent
                    same_parent_step_collection.steps_solution.append(
                        resolve_result.loaded_solution()
                    )
                    same_parent_step_collection.steps.append(step)

            else:  # add a step without collection (also parent)
                # put collection (if any) to queue
                if same_parent_step_collection.parent:
                    queue.put(
                        self._create_solution_run_collection_script(
                            same_parent_step_collection, script_creator
                        )
                    )

                # empty the collection for possible next steps
                same_parent_step_collection = SolutionGroup(step_solution_parsed_args)

                # run the old collection immediately
                if run_immediately:
                    self.run_queue(queue)

                # harvest arguments in the description of the step
                step_args = self._get_args(step, step_solution_parsed_args[0])

                # add step without parent
                queue.put(
                    self._create_solution_run_script_standalone(
                        resolve_result, step_args, script_creator
                    )
                )

        # put rest to queue
        if same_parent_step_collection.parent:
            queue.put(
                self._create_solution_run_collection_script(
                    same_parent_step_collection, script_creator
                )
            )

        self._print_credit(citations)

        # run the old collection immediately
        if run_immediately:
            self.run_queue(queue)

    def _create_solution_run_script_standalone(
        self,
        collection_solution: ICollectionSolution,
        args: list,
        script_creator: IScriptCreator,
    ) -> ScriptQueueEntry:
        """Creates the execution script for a album object giving its arguments.

        Args:
            script_creator:
                The ScriptCreator object to use to create the execution script.
            collection_solution:
                The collection solution object to create the executable script for.
            args:
                The arguments integrated in the script.

        Returns:
                The solution object and its scripts (in a list)

        """
        module_logger().debug(
            'Creating standalone album script "%s"...'
            % collection_solution.coordinates().name()
        )

        self._print_credit([collection_solution.loaded_solution()])
        environment = self.album.environment_manager().set_environment(
            collection_solution
        )
        script = script_creator.create_script(
            collection_solution.loaded_solution(), args
        )

        return ScriptQueueEntry(
            collection_solution.coordinates(), [script], environment=environment
        )

    def _create_solution_run_with_parent_script_standalone(
        self,
        collection_solution: ICollectionSolution,
        args: list,
        script_creator: ScriptCreatorRun,
    ) -> ScriptQueueEntry:
        """Creates the execution script for a album object having a parent dependency giving its arguments.

        Args:
            script_creator:
                The ScriptCreatorRun object to use to create the execution script.
            collection_solution:
                The collection solution object to create the executable script for.
            args:
                The arguments integrated in the script.

        Returns:
                The solution object and its scripts (in a list).

        """
        environment = self.album.environment_manager().set_environment(
            collection_solution
        )
        active_solution = collection_solution.loaded_solution()

        # create script
        script = script_creator.create_script(
            collection_solution.loaded_solution(), args
        )

        # TODO this should probably move into the runner
        self._print_credit([active_solution])

        return ScriptQueueEntry(
            collection_solution.loaded_solution().coordinates(), [script], environment
        )

    def _create_solution_run_collection_script(
        self, solution_collection: SolutionGroup, script_creator: ScriptCreatorRun
    ) -> ScriptQueueEntry:
        """Creates the execution script for a collection of solutions all having the same parent dependency.

        Args:
            script_creator:
                The ScriptCreator object to use to create the execution script.
            solution_collection
        Returns:
            The solution shared parent object and its scripts.

        """
        # load parent & steps
        parent_solution = self.album.state_manager().load(
            solution_collection.parent.path()
        )
        self.album.solutions().set_cache_paths(
            parent_solution, solution_collection.parent.catalog()
        )

        environment = self.album.environment_manager().set_environment(
            solution_collection.parent
        )
        module_logger().debug(
            'Creating script for steps (%s) with parent "%s"...'
            % (
                ", ".join(
                    [s.coordinates().name() for s in solution_collection.steps_solution]
                ),
                parent_solution.coordinates().name(),
            )
        )

        # handle arguments
        parsed_parent_args, parsed_steps_args_list = self._resolve_args(
            parent_solution=parent_solution,
            steps_solution=solution_collection.steps_solution,
            steps=solution_collection.steps,
            step_solution_parsed_args=solution_collection.parent_parsed_args,
        )

        # create script
        scripts = self._create_solution_run_with_parent_script(
            parent_solution,
            parsed_parent_args,
            solution_collection.steps_solution,
            parsed_steps_args_list,
            script_creator,
        )

        return ScriptQueueEntry(parent_solution.coordinates(), scripts, environment)

    @staticmethod
    def _create_solution_run_with_parent_script(
        parent_solution: ISolution,
        parent_args: list,
        child_solution_list: List[ISolution],
        child_args: list,
        script_creator: ScriptCreatorRun,
    ):
        """Creates the script for the parent solution as well as for all its steps (child solutions).

        Args:
            script_creator:
                The ScriptCreatorRun object to use to create the execution script.
            parent_solution:
                The parent solution object.
            parent_args:
                Arguments to use for the parent call.
            child_solution_list:
                A list of all solution objects to be executed with the same parent.
            child_args:
                List of arguments for the call of each child solution.

        Returns:
            A list holding all execution scripts.

        """
        module_logger().debug(
            'Creating album script with parent "%s"...'
            % parent_solution.coordinates().name()
        )

        script_creator_run_with_parent = ScriptCreatorRunWithParent(
            script_creator, child_solution_list, child_args
        )
        script = script_creator_run_with_parent.create_script(
            parent_solution, parent_args
        )
        script_list = [script]

        return script_list

    def __parse_args(self, active_solution: ISolution, args: list):
        """Parse arguments of loaded solution."""
        parser = argparse.ArgumentParser()

        class FileAction(argparse.Action):
            def __init__(self, option_strings, dest, nargs=None, **kwargs):
                if nargs is not None:
                    raise ValueError("nargs not allowed")
                super(FileAction, self).__init__(option_strings, dest, **kwargs)

            def __call__(self, p, namespace, values, option_string=None):
                setattr(
                    namespace,
                    self.dest,
                    active_solution.get_arg(self.dest)["action"](values),
                )

        for element in active_solution.setup()["args"]:
            if "action" in element.keys():
                parser.add_argument("--" + element["name"], action=FileAction)
            else:
                parser.add_argument("--" + element["name"])

        return parser.parse_known_args(args=args)

    def _resolve_args(
        self,
        parent_solution: ISolution,
        steps_solution: List[ISolution],
        steps: list,
        step_solution_parsed_args: list,
        args=None,
    ):
        """Resolves arguments of all steps and their parents."""
        args = [] if args is None else args
        parsed_parent_args = None
        parsed_steps_args_list = []

        module_logger().debug("Parsing arguments...")

        # iterate over all steps and parse arguments together
        for idx, step_solution in enumerate(steps_solution):
            step_parser = argparse.ArgumentParser()

            # the steps description
            step = steps[idx]

            if step:  # case steps argument resolving
                step_args = self._get_args(step, step_solution_parsed_args[0])
            else:  # case single step solution
                step_args = args

            # add parent arguments to the step album object arguments
            parent_dict = get_parent_dict(step_solution)
            if "args" in parent_dict:
                for param in parent_dict["args"]:
                    step_args.insert(0, f"--{param['name']}={str(param['value'])}")

            # add parent arguments
            if "args" in parent_solution.setup():
                [
                    step_parser.add_argument("--" + element["name"])
                    for element in parent_solution.setup()["args"]
                ]

            # parse all known arguments
            args_known, args_unknown = step_parser.parse_known_args(step_args)

            # only set parents args if not already set
            if not parsed_parent_args:
                parsed_parent_args = [""]
                parsed_parent_args.extend(
                    [
                        "--" + arg_name + "=" + getattr(args_known, arg_name)
                        for arg_name in vars(args_known)
                    ]
                )
                module_logger().debug(
                    'For step "%s" set parent arguments to %s...'
                    % (step_solution.coordinates().name(), parsed_parent_args)
                )

            # args_unknown are step args
            parsed_steps_args_list.append(args_unknown)
            module_logger().debug(
                'For step "%s" set step arguments to %s...'
                % (step_solution.coordinates().name(), args_unknown)
            )

        return parsed_parent_args, parsed_steps_args_list

    def _run_in_environment(self, script_queue_entry: ScriptQueueEntry):
        """Pushes a new logger to the stack before running the solution and pops it afterwards."""
        module_logger().debug(
            'Running script in environment of solution "%s"...'
            % script_queue_entry.coordinates.name()
        )
        self.album.environment_manager().run_scripts(
            script_queue_entry.environment, script_queue_entry.scripts
        )
        module_logger().debug(
            'Done running script in environment of solution "%s"...'
            % script_queue_entry.coordinates.name()
        )

    @staticmethod
    def _get_args(step, args):
        """Parse callable arguments belonging to a step into a list of strings"""
        argv = [""]
        if "args" in step:
            for param in step["args"]:
                argv.append(f"--{param['name']}={str(param['value'](args))}")
        return argv

    @staticmethod
    def _print_credit(active_solutions: List[ISolution]) -> None:
        res = ScriptManager._get_credit_as_string(active_solutions)
        if res:
            module_logger().info(res)

    @staticmethod
    def _get_credit_as_string(active_solutions: List[ISolution]):
        res = ""
        for active_solution in active_solutions:
            if active_solution.setup().cite:
                for citation in active_solution.setup().cite:
                    text = citation["text"]
                    if "doi" in citation:
                        text += " (DOI: %s)" % citation["doi"]
                    res += "%s\n" % text
        if len(res) > 0:
            return "\n\nSolution credits:\n\n%s\n" % res
        return None



from queue import Queue
from typing import List

import pkg_resources

from album.runner import album_logging

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.run_manager import IRunManager
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.model.default_values import DefaultValues
from album.core.model.event import Event
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.script_creator import ScriptCreatorRun

module_logger = album_logging.get_active_logger


class SolutionGroup:
    def __init__(
        self,
        parent_parsed_args=None,
        parent: ICollectionSolution = None,
        steps_solution=None,
        steps=None,
    ):
        if parent_parsed_args is None:
            parent_parsed_args = [None]
        if steps_solution is None:
            steps_solution = []
        if steps is None:
            steps = []
        self.parent_parsed_args = parent_parsed_args
        self.parent = parent
        self.steps_solution: List[
            ISolution
        ] = steps_solution  # The solution objects of all steps.

        self.steps = (
            steps  # The step description of the step. Must hold the argument keyword.
        )

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, SolutionGroup)
            and o.parent.coordinates() == self.parent.coordinates()
            and o.steps_solution == self.steps_solution
            and o.steps == self.steps
        )


class RunManager(IRunManager):
    def __init__(self, album: IAlbumController):
        self.album = album

        self.init_script = ""

    def run(self, solution_to_resolve: str, run_immediately=False, argv=None):
        """Run an already loaded solution."""
        resolve_result = self.album.collection_manager().resolve_installed_and_load(
            solution_to_resolve
        )
        coordinates = resolve_result.loaded_solution().coordinates()
        if not resolve_result.catalog:
            module_logger().debug("solution loaded locally: %s..." % str(coordinates))
        else:
            module_logger().debug(
                'solution loaded from catalog: "%s"...' % str(coordinates)
            )
        module_logger().debug('Initializing script to run "%s"' % coordinates)

        self.load_plugins(resolve_result)

        if argv is None:
            argv = [""]

        # pushing album and their scripts to a queue
        que = Queue()

        # builds the queue
        self.album.script_manager().build_queue(
            resolve_result, que, ScriptCreatorRun(), run_immediately, argv
        )

        self.album.event_manager().publish(
            Event(DefaultValues.before_run_event_name.value),
            resolve_result.coordinates(),
        )

        # runs the queue
        try:
            self.album.script_manager().run_queue(que)
        finally:
            self.album.event_manager().publish(
                Event(DefaultValues.after_run_event_name.value),
                resolve_result.coordinates(),
            )

        module_logger().debug(
            'Finished running script for solution "%s"' % coordinates.name()
        )

    def load_plugins(self, resolve_result):
        # process solution plugins
        module_logger().debug("Processing plugins...")
        if "dependencies" in resolve_result.loaded_solution().setup():
            deps = resolve_result.loaded_solution().setup()["dependencies"]
            if "plugins" in deps:
                module_logger().debug(
                    "Processing solution plugins: %s" % deps["plugins"]
                )
                available_plugins = pkg_resources.iter_entry_points("plugins_album")
                for plugin in deps["plugins"]:
                    found = False
                    for entry_point in available_plugins:
                        if entry_point.name == plugin["id"]:
                            module_logger().debug(
                                "Found plugin %s, initializing activation.."
                                % plugin["id"]
                            )
                            plugin_call = entry_point.load()
                            plugin_args = plugin["args"] if "args" in plugin else {}
                            from album.api import Album

                            plugin_call(
                                Album(self.album),
                                resolve_result.coordinates(),
                                plugin_args,
                            )
                            module_logger().debug("Plugin %s activated." % plugin["id"])
                            found = True
                            break
                    if not found:
                        raise LookupError(
                            "Cannot find plugin %s - please make sure the package providing this plugin "
                            "is installed into the album environment." % plugin["id"]
                        )
        module_logger().debug("Plugins processed.")



from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import List

from album.core.api.model.catalog import ICatalog
from album.runner.core.api.model.coordinates import ICoordinates


class ChangeType(Enum):
    ADDED = (1,)
    REMOVED = (2,)
    CHANGED = 3


class ISolutionChange:
    __metaclass__ = ABCMeta

    @abstractmethod
    def coordinates(self) -> ICoordinates:
        raise NotImplementedError

    @abstractmethod
    def change_type(self) -> ChangeType:
        raise NotImplementedError

    @abstractmethod
    def change_log(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def solution_status(self) -> dict:
        raise NotImplementedError

    @abstractmethod
    def as_dict(self) -> dict:
        raise NotImplementedError


class ICatalogAttributeChange:
    __metaclass__ = ABCMeta

    @abstractmethod
    def old_value(self):
        raise NotImplementedError

    @abstractmethod
    def new_value(self):
        raise NotImplementedError

    @abstractmethod
    def attribute(self) -> str:
        raise NotImplementedError


class ICatalogUpdates:
    __metaclass__ = ABCMeta

    @abstractmethod
    def catalog(self) -> ICatalog:
        raise NotImplementedError

    @abstractmethod
    def catalog_attribute_changes(self) -> List[ICatalogAttributeChange]:
        raise NotImplementedError

    @abstractmethod
    def solution_changes(self) -> List[ISolutionChange]:
        raise NotImplementedError

    @abstractmethod
    def as_dict(self) -> dict:
        raise NotImplementedError



from datetime import datetime
from pathlib import Path

from album.core.model.default_values import DefaultValues

from album.core.api.model.catalog import ICatalog
from album.core.utils.export.keepachangelog import to_raw_dict
from album.core.utils.operations.file_operations import create_path_recursively
from album.runner.album_logging import get_active_logger
from album.runner.core.api.model.solution import ISolution


def get_changelog_file_name():
    return DefaultValues.changelog_default_name.value


def get_changelog_content(
    active_solution: ISolution, catalog: ICatalog, dummy_content: str = None
):
    content = """# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),
and this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).
"""
    versions = catalog.get_all_solution_versions(
        active_solution.coordinates().group(), active_solution.coordinates().name()
    )
    for version in versions:
        timestamp = version.setup()["timestamp"]
        if timestamp:
            timestamp = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")
            time = timestamp.strftime("%Y-%m-%d")
            if (
                version.setup().version == active_solution.setup().version
                and dummy_content
            ):
                change = dummy_content
            else:
                if version.setup()["changelog"]:
                    change = version.setup()["changelog"]
                else:
                    change = ""
            content += "\n## [%s] - %s\n%s\n" % (
                version.setup()["version"],
                time,
                change,
            )
    return content


def create_changelog_file(
    active_solution: ISolution, catalog: ICatalog, target_folder: Path
):
    """Creates a changelog file in the given repo for the given solution.

    Returns:
        The Path to the created markdown file.

    """
    changelog_path = target_folder.joinpath(get_changelog_file_name())
    get_active_logger().debug("Writing changelog file to: %s..." % changelog_path)
    content = get_changelog_content(active_solution, catalog)

    create_path_recursively(changelog_path.parent)
    with open(str(changelog_path), "w+") as yml_f:
        yml_f.write(content)

    return changelog_path


def process_changelog_file(
    catalog: ICatalog, active_solution: ISolution, deploy_path: Path
):
    """Sets the changelog of a given solution."""
    changelog_name = get_changelog_file_name()
    changelog_file = Path(deploy_path).joinpath(changelog_name)

    if changelog_file.exists():
        # process existing changelog file
        changelogs = to_raw_dict(str(changelog_file))
        keyword = str(active_solution.setup().version).lower()
        if keyword in changelogs:
            active_solution.setup().changelog = changelogs[keyword]["raw"]
    else:
        # no changelog file found
        if not active_solution.setup().changelog:
            content = get_changelog_content(
                active_solution, catalog, "- INSERT LIST OF CHANGES"
            )
            get_active_logger().warning(
                "No %s file found.\nWe recommend documenting changes between versions. "
                "You can either\n\t- use the '--changelog' parameter of the deploy command\n\t"
                "- or add a file called %s next to the solution file.\nInsert what's printed "
                "between the following lines into %s and add your changes to the version you are "
                "about to release before running 'deploy':\n\n-----------------\n%s"
                "\n-----------------\n"
                % (changelog_name, changelog_name, changelog_name, content)
            )



from io import StringIO
from pathlib import Path

import validators

from album.core.api.model.environment import IEnvironment
from album.core.utils.operations.file_operations import (
    create_path_recursively,
    copy,
    write_dict_to_yml,
    get_dict_from_yml,
)
from album.core.utils.operations.url_operations import download_resource
from album.runner import album_logging

module_logger = album_logging.get_active_logger


class Environment(IEnvironment):
    def __init__(self, dependencies_dict, environment_name, cache_path: Path):
        """Init routine

        Args:
            dependencies_dict:
                Can be None or hold the entries a) "environment_file" b) "environment_name"
            environment_name:
                name prefix for all files to cache.
                Used when "name" is not available during yaml-file download for example.
            cache_path:
                cache path where to store downloads (yaml file, etc.)
        """
        self._name = environment_name
        self._cache_path = cache_path
        self._yaml_file = self._prepare_env_file(dependencies_dict)
        self._path = None

    def _prepare_env_file(self, dependencies_dict):
        """Checks how to set up an environment. Returns a path to a valid yaml file. Environment name in that file
        will be overwritten!

        Args:
            dependencies_dict:
                Dictionary holding the "environment_file" key. Environment file can be:
                    - url
                    - path
                    - stream object

        Returns:
            Path to a valid yaml file where environment name has been replaced!

        """
        if dependencies_dict:
            if "environment_file" in dependencies_dict:
                env_file = dependencies_dict["environment_file"]

                yaml_path = self._cache_path.joinpath("%s%s" % (self._name, ".yml"))
                create_path_recursively(yaml_path.parent)

                if isinstance(env_file, str):
                    # case valid url
                    if validators.url(env_file):
                        yaml_path = download_resource(env_file, yaml_path)
                    # case file content
                    elif "dependencies:" in env_file and "\n" in env_file:
                        with open(str(yaml_path), "w+") as f:
                            f.writelines(env_file)
                        yaml_path = yaml_path
                    # case Path
                    elif Path(env_file).is_file() and Path(env_file).stat().st_size > 0:
                        yaml_path = copy(env_file, yaml_path)
                    else:
                        raise TypeError(
                            "environment_file must either contain the content of the environment file, "
                            "contain the url to a valid file or point to a file on the disk!"
                        )
                # case String stream
                elif isinstance(env_file, StringIO):
                    with open(str(yaml_path), "w+") as f:
                        env_file.seek(0)  # make sure we start from the beginning
                        f.writelines(env_file.readlines())
                    yaml_path = yaml_path
                else:
                    raise RuntimeError(
                        "Environment file specified, but format is unknown!"
                        " Don't know where to run solution!"
                    )

                yaml_dict = get_dict_from_yml(yaml_path)
                yaml_dict["name"] = self._name
                write_dict_to_yml(yaml_path, yaml_dict)

                return yaml_path
            return None
        return None

    def name(self):
        return self._name

    def cache_path(self):
        return self._cache_path

    def yaml_file(self):
        return self._yaml_file

    def path(self):
        return self._path

    def set_path(self, path):
        self._path = path



from pathlib import Path
from typing import Optional, List

from album.core.api.controller.collection.catalog_handler import ICatalogHandler
from album.core.api.controller.collection.collection_manager import ICollectionManager
from album.core.api.controller.collection.solution_handler import ISolutionHandler
from album.core.api.controller.controller import IAlbumController
from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.controller.collection.catalog_handler import CatalogHandler
from album.core.controller.collection.solution_handler import SolutionHandler
from album.core.model.collection_index import CollectionIndex
from album.core.model.default_values import DefaultValues
from album.core.model.resolve_result import ResolveResult
from album.core.utils.operations.file_operations import write_dict_to_json
from album.core.utils.operations.resolve_operations import (
    dict_to_coordinates,
    get_attributes_from_string,
    check_doi,
    get_doi_from_input,
    check_file_or_url,
)
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates

module_logger = album_logging.get_active_logger


class CollectionManager(ICollectionManager):
    def __init__(self, album: IAlbumController):
        self.album = album
        self.solution_handler = SolutionHandler(self.album)
        self.catalog_handler = CatalogHandler(self.album)
        self.catalog_collection: Optional[ICollectionIndex] = None
        self.collection_loaded = False

    def __del__(self):
        self.close()

    def close(self):
        if self.solution_handler is not None:
            del self.solution_handler
            self.solution_handler = None
        if self.catalog_handler is not None:
            del self.catalog_handler
            self.catalog_handler = None
        if self.catalog_collection is not None:
            self.catalog_collection.close()
            self.catalog_collection = None

    def get_collection_index(self) -> ICollectionIndex:
        return self.catalog_collection

    def load_or_create(self):
        if self.collection_loaded:
            module_logger().warning(
                "CollectionManager().load_or_create_collection() should only be called once."
            )
            return
        self.collection_loaded = True
        collection_meta = self.album.configuration().get_catalog_collection_meta_dict()
        newly_created = (
            not self.album.configuration().get_catalog_collection_path().exists()
        )
        if collection_meta:
            collection_version = collection_meta["catalog_collection_version"]
        else:
            if newly_created:
                collection_version = CollectionIndex.version
                self.write_version_to_yml(
                    self.album.configuration().get_catalog_collection_meta_path(),
                    "my-collection",
                    collection_version,
                )
            else:
                raise RuntimeError(
                    "Album collection database file found, but no meta file specifying the database version."
                )
        if self.catalog_collection is not None:
            self.catalog_collection.close()
        self.catalog_collection = CollectionIndex(
            name=DefaultValues.catalog_collection_name.value,
            path=self.album.configuration().get_catalog_collection_path(),
        )
        self.album.migration_manager().migrate_collection_index(
            self.catalog_collection, initial_version=collection_version
        )
        if newly_created:
            self.catalog_handler.add_initial_catalogs()

    def catalogs(self) -> ICatalogHandler:
        return self.catalog_handler

    def solutions(self) -> ISolutionHandler:
        return self.solution_handler

    def get_index_as_dict(self):
        catalogs = self.catalog_collection.get_all_catalogs()
        for catalog in catalogs:
            solutions = []
            for solution in self.catalog_collection.get_solutions_by_catalog(
                catalog["catalog_id"]
            ):
                solutions.append(
                    {"setup": solution.setup(), "internal": solution.internal()}
                )
            catalog["solutions"] = solutions
        return {
            "base": str(self.album.configuration().base_cache_path()),
            "catalogs": catalogs,
        }

    def resolve_installed(self, resolve_solution) -> ICollectionSolution:
        resolve_result = self._resolve(resolve_solution)

        if not resolve_result.database_entry():
            raise LookupError("Solution not found!")

        if not resolve_result.database_entry().internal()["installed"]:
            raise ValueError(
                "Solution seems not to be installed! Please install solution first!"
            )

        return resolve_result

    def resolve_installed_and_load(self, resolve_solution) -> ICollectionSolution:
        resolve_result = self.resolve_installed(resolve_solution)

        loaded_solution = self.album.state_manager().load(resolve_result.path())
        self.solution_handler.set_cache_paths(loaded_solution, resolve_result.catalog())

        resolve_result.set_loaded_solution(loaded_solution)

        return resolve_result

    def resolve_and_load(self, resolve_solution) -> ICollectionSolution:
        resolve_result = self._resolve(resolve_solution)
        self.retrieve_and_load_resolve_result(resolve_result)
        self.solution_handler.set_cache_paths(
            resolve_result.loaded_solution(), resolve_result.catalog()
        )

        return resolve_result

    def resolve_and_load_catalog_coordinates(
        self, catalog: ICatalog, coordinates: ICoordinates
    ) -> ICollectionSolution:
        collection_entry = self._search_in_specific_catalog(
            catalog.catalog_id(), coordinates
        )
        solution_path = self.solution_handler.get_solution_file(catalog, coordinates)

        resolve_result = ResolveResult(
            path=solution_path,
            catalog=catalog,
            collection_entry=collection_entry,
            coordinates=coordinates,
        )
        self.retrieve_and_load_resolve_result(resolve_result)

        self.solution_handler.set_cache_paths(
            resolve_result.loaded_solution(), resolve_result.catalog()
        )

        return resolve_result

    def resolve_and_load_coordinates(
        self, coordinates: ICoordinates
    ) -> ICollectionSolution:
        collection_entry = self._search_by_coordinates(coordinates)
        catalog = self.album.catalogs().get_by_id(
            collection_entry.internal()["catalog_id"]
        )

        solution_path = self.solution_handler.get_solution_file(catalog, coordinates)
        resolve_result = ResolveResult(
            path=solution_path,
            catalog=catalog,
            collection_entry=collection_entry,
            coordinates=coordinates,
        )
        self.retrieve_and_load_resolve_result(resolve_result)

        self.solution_handler.set_cache_paths(
            resolve_result.loaded_solution(), resolve_result.catalog()
        )

        return resolve_result

    def resolve(self, resolve_solution) -> ICollectionSolution:
        resolve_result = self._resolve(resolve_solution)

        if not Path(resolve_result.path()).exists():
            self.solution_handler.retrieve_solution(
                resolve_result.catalog(), resolve_result.coordinates()
            )

        return resolve_result

    def _resolve(self, str_input: str) -> ICollectionSolution:

        str_input = str(str_input)

        # always first resolve outside any catalog, excluding a DOI which should be first resolved inside a catalog
        path = check_file_or_url(
            str_input, self.album.configuration().cache_path_download()
        )

        doi = get_doi_from_input(str_input)
        if path:
            # will load the solution behind the path to get meta-information
            solution_entry = self._search_for_local_file(path)

            # a solution loaded this way will always end up in a local catalog
            catalog = self.album.catalogs().get_cache_catalog()
        else:
            # search DOI first
            if doi:
                solution_entry = self._search_doi(doi["doi"])

                # either a doi is found in the collection or it will be downloaded and ends up in a local catalog
                if solution_entry:
                    catalog = self.album.catalogs().get_by_id(
                        solution_entry.internal()["catalog_id"]
                    )
                    path = self.solution_handler.get_solution_file(
                        catalog, dict_to_coordinates(solution_entry.setup())
                    )
                else:
                    # download DOI
                    path = check_doi(doi["doi"], self.album.configuration().tmp_path())
                    catalog = self.album.catalogs().get_cache_catalog()
            else:  # case no doi
                solution_entry = self._search(str_input)

                if not solution_entry:
                    raise LookupError(
                        "Cannot find solution %s! Try <doi>:<prefix>/<suffix> or <prefix>/<suffix> "
                        "or <group>:<name>:<version> or <catalog>:<group>:<name>:<version> "
                        "or point to a valid file or folder! Aborting..." % str_input
                    )

                catalog = self.album.catalogs().get_by_id(
                    solution_entry.internal()["catalog_id"]
                )
                path = self.solution_handler.get_solution_file(
                    catalog, dict_to_coordinates(solution_entry.setup())
                )

        coordinates = None

        if solution_entry:
            coordinates = dict_to_coordinates(solution_entry.setup())

        resolve = ResolveResult(
            path=path,
            catalog=catalog,
            collection_entry=solution_entry,
            coordinates=coordinates,
        )

        return resolve

    def _search_for_local_file(self, path) -> ICollectionIndex.ICollectionSolution:
        active_solution = self.album.state_manager().load(path)

        # check in collection
        solution_entry = (
            self.catalog_collection.get_solution_by_catalog_grp_name_version(
                self.album.catalogs().get_cache_catalog().catalog_id(),
                active_solution.coordinates(),
            )
        )

        return solution_entry

    def _search_doi(self, doi):
        solution_entry = self.catalog_collection.get_solution_by_doi(doi)

        return solution_entry

    def _search(self, str_input) -> ICollectionIndex.ICollectionSolution:
        """Searches ONLY in the catalog collection, given a string which is interpreted."""
        try:
            attrs = get_attributes_from_string(str_input)
        except ValueError:
            return self._guess(str_input)
        solution_entry = None
        if "doi" in attrs:  # case doi
            solution_entry = self._search_doi(attrs["doi"])
        else:
            coordinates = dict_to_coordinates(attrs)
            if "catalog" not in attrs:
                solution_entry = self._search_in_local_catalog(
                    coordinates
                )  # resolve in local catalog first!

            if not solution_entry:
                if "catalog" in attrs:  # resolve in specific catalog
                    catalog_id = (
                        self.album.catalogs().get_by_name(attrs["catalog"]).catalog_id()
                    )
                    solution_entry = self._search_in_specific_catalog(
                        catalog_id, coordinates
                    )
                else:
                    solution_entries = self._search_in_catalogs(
                        coordinates
                    )  # resolve anywhere

                    if solution_entries and len(solution_entries) > 1:
                        module_logger().warning(
                            "Found several solutions... taking the first one! "
                        )

                    if solution_entries:
                        solution_entry = solution_entries[0]
        return solution_entry

    def _search_by_coordinates(
        self, coordinates: ICoordinates
    ) -> Optional[ICollectionIndex.ICollectionSolution]:
        solution_entry = self._search_in_local_catalog(
            coordinates
        )  # resolve in local catalog first!

        if not solution_entry:
            solution_entries = self._search_in_catalogs(coordinates)  # resolve anywhere

            if solution_entries and len(solution_entries) > 1:
                module_logger().warning(
                    "Found several solutions... taking the first one! "
                )

            if solution_entries:
                solution_entry = solution_entries[0]

        return solution_entry

    def _search_in_local_catalog(
        self, coordinates: ICoordinates
    ) -> Optional[ICollectionIndex.ICollectionSolution]:
        """Searches in the local catalog only"""
        return self._search_in_specific_catalog(
            self.album.catalogs().get_cache_catalog().catalog_id(), coordinates
        )

    def _search_in_specific_catalog(
        self, catalog_id, coordinates: ICoordinates
    ) -> Optional[ICollectionIndex.ICollectionSolution]:
        """Searches in a given catalog only"""
        return self.catalog_collection.get_solution_by_catalog_grp_name_version(
            catalog_id, coordinates
        )

    def _search_in_catalogs(
        self, coordinates: ICoordinates
    ) -> List[ICollectionIndex.ICollectionSolution]:
        """Searches the whole collection giving coordinates"""
        solution_entries = self.catalog_collection.get_solutions_by_grp_name_version(
            coordinates
        )

        return solution_entries if solution_entries else None

    def retrieve_and_load_resolve_result(self, resolve_result: ICollectionSolution):
        if not resolve_result.path() or not Path(resolve_result.path()).exists():
            self.solution_handler.retrieve_solution(
                resolve_result.catalog(),
                dict_to_coordinates(resolve_result.database_entry().setup()),
            )
        resolve_result.set_loaded_solution(
            self.album.state_manager().load(resolve_result.path())
        )
        resolve_result.set_coordinates(resolve_result.loaded_solution().coordinates())

    @staticmethod
    def write_version_to_yml(path, name, version) -> None:
        write_dict_to_json(
            path,
            {"catalog_collection_name": name, "catalog_collection_version": version},
        )

    def _guess(self, str_input) -> Optional[ICollectionIndex.ICollectionSolution]:
        input_parts = str_input.split(":")
        call_not_reproducible = (
            "This call is not fully reproducible. Resolving to this solution: %s"
        )

        if len(input_parts) == 1:
            solutions = self.catalog_collection.get_solutions_by_name(input_parts[0])
            if len(solutions) == 1:
                module_logger().warn(
                    call_not_reproducible % dict_to_coordinates(solutions[0].setup())
                )
                return solutions[0]
            if len(solutions) > 1:
                return self._handle_multiple_solution_matches(solutions)
            else:
                return None
        if len(input_parts) == 2:
            solutions = self.catalog_collection.get_solutions_by_name_version(
                input_parts[0], input_parts[1]
            )
            if len(solutions) == 1:
                module_logger().warn(
                    call_not_reproducible % dict_to_coordinates(solutions[0].setup())
                )
                return solutions[0]
            if len(solutions) > 1:
                return self._handle_multiple_solution_matches(solutions)
            if len(solutions) == 0:
                solutions = self.catalog_collection.get_solutions_by_grp_name(
                    input_parts[0], input_parts[1]
                )
                if len(solutions) == 1:
                    module_logger().warn(
                        call_not_reproducible
                        % dict_to_coordinates(solutions[0].setup())
                    )
                    return solutions[0]
                if len(solutions) > 1:
                    return self._handle_multiple_solution_matches(solutions)
        return None

    def _handle_multiple_solution_matches(
        self, solutions: [ICollectionIndex.ICollectionSolution]
    ):
        call_not_reproducible = "Resolving ambiguous input to %s"
        cache_id = self.catalogs().get_cache_catalog().catalog_id()
        cache_matches = [
            solution
            for solution in solutions
            if solution.internal()["catalog_id"] == cache_id
        ]
        if len(cache_matches) == 1:
            module_logger().warn(
                call_not_reproducible % dict_to_coordinates(cache_matches[0].setup())
            )
            return cache_matches[0]
        solutions_ambiguous_str = (
            "Input is ambiguous - choose between one of these solutions:\n%s"
        )
        raise RuntimeError(solutions_ambiguous_str % self._solutions_as_list(solutions))

    def _solutions_as_list(self, solutions: [ICollectionIndex.ICollectionSolution]):
        solutions_str = ""
        for solution in solutions:
            catalog_name = (
                self.catalogs().get_by_id(solution.internal()["catalog_id"]).name()
            )
            solutions_str += "- %s:%s\n" % (
                catalog_name,
                dict_to_coordinates(solutions[0].setup()),
            )
        return solutions_str



from typing import Optional

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.state_manager import IStateManager
from album.runner import api
from album.runner.album_logging import get_active_logger
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.solution import Solution


class StateManager(IStateManager):
    def __init__(self, album: IAlbumController):
        self._active_solution = []
        self.album = album
        # overwrite album setup with this setup
        api.setup = self._setup_solution

    def load(self, path) -> Optional[ISolution]:
        get_active_logger().debug(f"Loading solution from {path}...")
        with open(path, "r") as f:
            solution = f.read()
        exec(solution)
        active_solution = self._get_active_solution()
        active_solution.set_script(solution)
        self._pop_active_solution()
        return active_solution

    def _setup_solution(self, **attrs):
        """This configures a solution for use by the main album tool."""
        attrs = self.album.migration_manager().migrate_solution_attrs(attrs)
        next_solution = Solution(attrs)
        self._push_active_solution(next_solution)

    def _push_active_solution(self, solution_object: ISolution):
        """Pop a album to the _active_solution stack."""
        self._active_solution.insert(0, solution_object)

    def _get_active_solution(self) -> Optional[ISolution]:
        """Return the currently active solution, which is defined globally."""
        if len(self._active_solution) > 0:
            return self._active_solution[0]
        return None

    def _pop_active_solution(self):
        """Pop the currently active album from the _active_solution stack."""
        if len(self._active_solution) > 0:
            return self._active_solution.pop(0)
        else:
            return None



import album
from album.ci.argument_parsing import main
from album.core.utils.core_logging import configure_root_logger
from album.runner import album_logging
from album.runner.album_logging import get_active_logger, debug_settings

module_logger = get_active_logger


def startup():
    __retrieve_logger()
    module_logger().info(
        "Running album-catalog-admin version %s \n \n %s - contact via %s "
        % (album.ci.__version__, album.ci.__author__, album.ci.__email__)
    )
    main()


def __retrieve_logger():
    """Retrieves the default album ci logger."""
    configure_root_logger(album_logging.LogLevel(debug_settings()))


if __name__ == "__main__":
    startup()



import pkgutil
from datetime import datetime
from typing import Optional, List

from album.core.api.model.catalog_index import ICatalogIndex
from album.core.model.database import Database
from album.core.utils.operations.file_operations import (
    get_dict_entry,
    write_dict_to_json,
)
from album.core.utils.operations.solution_operations import (
    get_solution_hash,
    serialize_json,
)
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates

module_logger = album_logging.get_active_logger


class CatalogIndex(ICatalogIndex, Database):
    version = "0.1.0"

    def __init__(self, name, path):
        """Init routine.

        Args:
            name:
                The name of the catalog.
            path:
                The path to the index file.

        """
        self.name = name
        super().__init__(path)

    def create(self):
        data = pkgutil.get_data("album.core.schema", "catalog_index_schema.sql")

        cursor = self.get_cursor()
        cursor.executescript(data.decode("utf-8"))

        self.update_name_version(self.name, self.version, close=False)

        # explicitly commit and close
        self.close_current_connection(commit=True)

    def is_empty(self, close=True):
        cursor = self.get_cursor()
        r = cursor.execute("SELECT * FROM solution").fetchone()

        if close:
            self.close_current_connection()

        return False if r else True

    # ### catalog_index ###

    def update_name_version(self, name, version, close=True):
        module_logger().debug(
            'Update index name: "%s" and version: "%s"' % (name, version)
        )

        curr_name = self.get_name()

        cursor = self.get_cursor()
        if curr_name:
            cursor.execute(
                "UPDATE catalog_index SET name=:name, version=:version WHERE name_id=:name_id",
                {"name_id": 1, "name": name, "version": version},
            )
        else:
            cursor.execute(
                "INSERT INTO catalog_index values (?, ?, ?)", (1, name, version)
            )

        if close:
            self.close_current_connection()

    def get_name(self, close=True):
        module_logger().debug("Get catalog name...")

        cursor = self.get_cursor()
        r = cursor.execute("SELECT * FROM catalog_index").fetchone()

        cur_name = r["name"] if r else None

        if close:
            self.close_current_connection()

        return cur_name

    def get_version(self, close=True):
        module_logger().debug("Get index version...")

        cursor = self.get_cursor()
        r = cursor.execute("SELECT * FROM catalog_index").fetchone()

        cur_version = r["version"] if r else None

        if close:
            self.close_current_connection()

        return cur_version

    def get_all_solutions(self, close=True):
        module_logger().debug("Retrieve all solutions...")

        cursor = self.get_cursor()
        r = cursor.execute("SELECT * FROM solution", {}).fetchall()

        solutions = []
        if r:
            for s in r:
                solution = dict(s)
                self._append_metadata_to_solution_dict(solution)
                solutions.append(solution)

        if close:
            self.close_current_connection()

        return solutions

    @staticmethod
    def get_solution_column_keys():
        return [
            "group",
            "name",
            "description",
            "version",
            "album_api_version",
            "album_version",
            "license",
            "acknowledgement",
            "title",
            "timestamp",
            "doi",
            "changelog",
        ]

    def _insert_solution(self, solution_attrs, close=True) -> int:
        hash_val = get_solution_hash(solution_attrs, self.get_solution_column_keys())
        solution_id = self.next_id("solution")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO solution values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            (
                solution_id,
                solution_attrs["group"],
                solution_attrs["name"],
                get_dict_entry(solution_attrs, "title"),
                solution_attrs["version"],
                datetime.now().isoformat(),
                get_dict_entry(solution_attrs, "description"),
                get_dict_entry(solution_attrs, "doi"),  # allow to be none
                get_dict_entry(solution_attrs, "license"),
                get_dict_entry(solution_attrs, "album_version"),
                get_dict_entry(solution_attrs, "album_api_version"),
                get_dict_entry(solution_attrs, "changelog"),  # allow to be none
                get_dict_entry(solution_attrs, "acknowledgement"),
                hash_val,
            ),
        )
        if "solution_creators" in solution_attrs:
            for author in solution_attrs["solution_creators"]:
                author_id = self._exists_author(author, close=False)
                if not author_id:
                    author_id = self._insert_author(author, close=False)
                solution_author_id = self.next_id("solution_author")
                cursor.execute(
                    "INSERT INTO solution_author values (?, ?, ?)",
                    (solution_author_id, solution_id, author_id),
                )

        if "tags" in solution_attrs:
            for tag in solution_attrs["tags"]:
                tag_id = self._exists_tag(tag, close=False)
                if not tag_id:
                    tag_id = self._insert_tag(tag, close=False)
                solution_tag_id = self.next_id("solution_tag")
                cursor.execute(
                    "INSERT INTO solution_tag values (?, ?, ?)",
                    (solution_tag_id, solution_id, tag_id),
                )

        if "args" in solution_attrs:
            for argument in solution_attrs["args"]:
                argument_id = self._exists_argument(argument, close=False)
                if not argument_id:
                    argument_id = self._insert_argument(argument, close=False)
                solution_argument_id = self.next_id("solution_argument")
                cursor.execute(
                    "INSERT INTO solution_argument values (?, ?, ?)",
                    (solution_argument_id, solution_id, argument_id),
                )
        if "cite" in solution_attrs:
            for citation in solution_attrs["cite"]:
                citation_id = self._exists_citation(citation, close=False)
                if not citation_id:
                    citation_id = self._insert_citation(citation, close=False)
                solution_citation_id = self.next_id("solution_citation")
                cursor.execute(
                    "INSERT INTO solution_citation values (?, ?, ?)",
                    (solution_citation_id, solution_id, citation_id),
                )

        if "custom" in solution_attrs:
            for key in solution_attrs["custom"]:
                value = solution_attrs["custom"][key]
                custom_id = self._exists_custom_key(key, value, close=False)
                if not custom_id:
                    custom_id = self._insert_custom_key(key, value, close=False)
                solution_custom_key_id = self.next_id("solution_custom")
                cursor.execute(
                    "INSERT INTO solution_custom values (?, ?, ?)",
                    (solution_custom_key_id, solution_id, custom_id),
                )

        if "covers" in solution_attrs:
            for cover in solution_attrs["covers"]:
                if not self._exists_cover(cover, solution_id):
                    self._insert_cover(cover, solution_id, close=False)

        if "documentation" in solution_attrs:
            for documentation in solution_attrs["documentation"]:
                if not self._exists_documentation(documentation, solution_id):
                    self._insert_documentation(documentation, solution_id, close=False)

        self.save()

        if close:
            self.close_current_connection()

        return solution_id

    def _exists_author(self, author_name, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM author WHERE name=:author_name", {"author_name": author_name}
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["author_id"] if r else None

    def _insert_author(self, author, close=True):
        author_id = self.next_id("author")

        cursor = self.get_cursor()
        cursor.execute("INSERT INTO author values (?, ?)", (author_id, author))

        if close:
            self.close_current_connection()

        return author_id

    def _exists_tag(self, tag_name, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM tag WHERE name=:tag_name ", {"tag_name": tag_name}
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["tag_id"] if r else None

    def _insert_tag(self, tag, close=True):
        tag_id = self.next_id("tag")

        cursor = self.get_cursor()
        cursor.execute("INSERT INTO tag values (?, ?, ?)", (tag_id, tag, "manual"))

        if close:
            self.close_current_connection()

        return tag_id

    def _exists_argument(self, argument, close=True):
        cursor = self.get_cursor()

        exc_str = "SELECT * FROM argument WHERE name=:argument_name AND description=:argument_description "
        exc_val = {
            "argument_name": argument["name"],
            "argument_description": argument["description"],
        }
        argument_type = get_dict_entry(argument, "type")
        argument_default_value = get_dict_entry(argument, "default")

        if argument_type:
            exc_str += "AND type=:argument_type "
            exc_val["argument_type"] = argument_type
        else:
            exc_str += "AND type IS NULL "

        if argument_default_value:
            exc_str += "AND default_value=:argument_default_value "
            exc_val["argument_default_value"] = argument_default_value
        else:
            exc_str += "AND default_value IS NULL "

        r = cursor.execute(exc_str, exc_val).fetchone()

        if close:
            self.close_current_connection()

        return r["argument_id"] if r else None

    def _exists_custom_key(self, custom_key, custom_value, close=True):
        cursor = self.get_cursor()

        exc_str = "SELECT * FROM custom WHERE custom_key=:custom_key AND custom_value=:custom_value "
        exc_val = {
            "custom_key": custom_key,
            "custom_value": custom_value,
        }

        r = cursor.execute(exc_str, exc_val).fetchone()

        if close:
            self.close_current_connection()

        return r["custom_id"] if r else None

    def _insert_argument(self, argument, close=True):
        argument_id = self.next_id("argument")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO argument values (?, ?, ?, ?, ?, ?)",
            (
                argument_id,
                argument["name"],
                get_dict_entry(argument, "type"),
                get_dict_entry(argument, "description"),
                get_dict_entry(argument, "default"),
                get_dict_entry(argument, "required"),
            ),
        )

        if close:
            self.close_current_connection()

        return argument_id

    def _insert_custom_key(self, custom_key, custom_value, close=True):
        custom_id = self.next_id("custom")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO custom values (?, ?, ?)", (custom_id, custom_key, custom_value)
        )

        if close:
            self.close_current_connection()

        return custom_id

    def _exists_citation(self, citation, close=True):
        cursor = self.get_cursor()

        exc_str = "SELECT * FROM citation WHERE text=:citation_text "
        exc_val = {
            "citation_text": citation["text"],
        }
        citation_doi = get_dict_entry(citation, "doi")

        if citation_doi:
            exc_str += "AND doi=:citation_doi"
            exc_val["citation_doi"] = citation_doi
        else:
            exc_str += "AND doi IS NULL"

        r = cursor.execute(exc_str, exc_val).fetchone()

        if close:
            self.close_current_connection()

        return r["citation_id"] if r else None

    def _insert_citation(self, citation, close=True):
        citation_id = self.next_id("citation")
        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO citation values (?, ?, ?, ?)",
            (
                citation_id,
                citation["text"],
                get_dict_entry(citation, "doi"),
                get_dict_entry(citation, "url"),
            ),
        )

        if close:
            self.close_current_connection()

        return citation_id

    def _exists_cover(self, cover, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM cover WHERE source=:cover_source AND description=:cover_description "
            "AND solution_id=:solution_id",
            {
                "cover_source": cover["source"],
                "cover_description": cover["description"],
                "solution_id": solution_id,
            },
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["cover_id"] if r else None

    def _insert_cover(self, cover, solution_id, close=True):
        cover_id = self.next_id("cover")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO cover values (?, ?, ?, ?)",
            (cover_id, solution_id, cover["source"], cover["description"]),
        )

        if close:
            self.close_current_connection()

        return cover_id

    def _exists_documentation(self, documentation, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM main.documentation WHERE documentation=:documentation "
            "AND solution_id=:solution_id",
            {
                "documentation": documentation,
                "solution_id": solution_id,
            },
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["documentation_id"] if r else None

    def _insert_documentation(self, documentation, solution_id, close=True):
        documentation_id = self.next_id("documentation")
        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO documentation values (?, ?, ?)",
            (documentation_id, solution_id, documentation),
        )

        if close:
            self.close_current_connection()

        return documentation_id

    def get_solution(self, solution_id, close=True) -> Optional[dict]:
        module_logger().debug('Get solution by id: "%s"...' % solution_id)

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM solution WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id,
            },
        ).fetchone()
        solution = None
        if r:
            solution = dict(r)
            self._append_metadata_to_solution_dict(solution)

        if close:
            self.close_current_connection()

        return solution

    def get_solution_by_coordinates(
        self, coordinates: ICoordinates, close=True
    ) -> Optional[dict]:
        module_logger().debug('Get solution by coordinates: "%s"...' % str(coordinates))

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT s.* FROM solution s "
            'WHERE s."group"=:group AND s.name=:name AND s.version=:version',
            {
                "group": coordinates.group(),
                "name": coordinates.name(),
                "version": coordinates.version(),
            },
        ).fetchone()

        solution = None
        if r:
            solution = dict(r)
            self._append_metadata_to_solution_dict(solution)

        if close:
            self.close_current_connection()

        return solution

    def _append_metadata_to_solution_dict(self, solution):
        solution_id = solution["solution_id"]
        solution["solution_creators"] = self._get_authors_by_solution(solution_id)
        solution["covers"] = self._get_covers_by_solution(solution_id)
        solution["documentation"] = self._get_documentation_by_solution(solution_id)
        solution["args"] = self._get_arguments_by_solution(solution_id)
        solution["cite"] = self._get_citations_by_solution(solution_id)
        solution["tags"] = self._get_tags_by_solution(solution_id)
        solution["custom"] = self._get_custom_by_solution(solution_id)

    def get_solution_by_doi(self, doi, close=True) -> Optional[dict]:
        module_logger().debug('Get solution by doi: "%s"...' % doi)

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM solution WHERE doi=:doi_value",
            {
                "doi_value": doi,
            },
        ).fetchone()

        solution = None
        if r:
            solution = dict(r)

        if close:
            self.close_current_connection()

        return solution

    def get_all_solution_versions(
        self, group, name, close=True
    ) -> Optional[List[dict]]:
        module_logger().debug(
            'Get solution by group and name: "%s:%s"...' % (group, name)
        )

        cursor = self.get_cursor()
        r = cursor.execute(
            'SELECT * FROM solution WHERE "group"=:group_value AND name=:name_value ORDER BY version DESC',
            {
                "group_value": group,
                "name_value": name,
            },
        ).fetchall()

        solutions = []
        if r:
            for solution in r:
                solutions.append(dict(solution))

        if close:
            self.close_current_connection()

        return solutions

    def _update_solution(
        self, coordinates: ICoordinates, solution_attrs: dict, close=True
    ) -> None:
        try:
            # it is easier to delete and insert again instead of updating all connection-tables
            solution_dict = self.get_solution_by_coordinates(coordinates, close=False)
            if solution_dict:
                self.remove_solution(solution_dict["solution_id"])
            else:
                raise RuntimeError(
                    'Cannot update solution "%s"! Does not exist!' % str(coordinates)
                )

            self._insert_solution(solution_attrs, close=False)

            self.save()
        finally:
            if close:
                self.close_current_connection()

    def remove_solution(self, solution_id, close=True):
        cursor = self.get_cursor()

        cursor.execute(
            "DELETE FROM cover WHERE solution_id=:solution_id",
            {"solution_id": solution_id},
        )
        cursor.execute(
            "DELETE FROM documentation WHERE solution_id=:solution_id",
            {"solution_id": solution_id},
        )

        cursor.execute(
            "DELETE FROM solution_tag WHERE solution_id=:solution_id",
            {"solution_id": solution_id},
        )

        cursor.execute(
            "DELETE FROM solution_author WHERE solution_id=:solution_id",
            {"solution_id": solution_id},
        )

        cursor.execute(
            "DELETE FROM solution_citation WHERE solution_id=:solution_id",
            {"solution_id": solution_id},
        )

        cursor.execute(
            "DELETE FROM solution_argument WHERE solution_id=:solution_id",
            {"solution_id": solution_id},
        )

        cursor.execute(
            "DELETE FROM solution_custom WHERE solution_id=:solution_id",
            {"solution_id": solution_id},
        )

        cursor.execute(
            "DELETE FROM tag "
            "WHERE NOT EXISTS (SELECT st.tag_id FROM solution_tag st "
            "WHERE tag.tag_id = st.tag_id)"
        )

        cursor.execute(
            "DELETE FROM argument "
            "WHERE NOT EXISTS (SELECT sa.argument_id FROM solution_argument sa "
            "WHERE argument.argument_id = sa.argument_id)"
        )

        cursor.execute(
            "DELETE FROM citation "
            "WHERE NOT EXISTS (SELECT sc.citation_id FROM solution_citation sc "
            "WHERE citation.citation_id = sc.citation_id)"
        )

        cursor.execute(
            "DELETE FROM author "
            "WHERE NOT EXISTS (SELECT sa.author_id FROM solution_author sa "
            "WHERE author.author_id = sa.author_id)"
        )

        cursor.execute(
            "DELETE FROM custom "
            "WHERE NOT EXISTS (SELECT sa.custom_id FROM solution_custom sa "
            "WHERE custom.custom_id = sa.custom_id)"
        )

        cursor.execute(
            "DELETE FROM solution WHERE solution_id=:solution_id",
            {"solution_id": solution_id},
        )

        self.save()

        if close:
            self.close_current_connection()

    def remove_solution_by_group_name_version(
        self, coordinates: ICoordinates, close=True
    ):
        solution_dict = self.get_solution_by_coordinates(coordinates, close=close)
        if solution_dict:
            self.remove_solution(solution_dict["solution_id"])
            return solution_dict
        return None

    # ### catalog_features ###

    def update(self, coordinates: ICoordinates, solution_attrs: dict, close=True):
        if self.get_solution_by_coordinates(coordinates, close=False):
            module_logger().debug("Update solution...")
            self._update_solution(coordinates, solution_attrs, close=close)
        else:
            module_logger().debug("Insert solution...")
            self._insert_solution(solution_attrs, close=close)

    def save(self):
        module_logger().debug("Saving index...")
        self.get_connection().commit()

    def export(self, path, export_format="JSON", close=True):
        module_logger().debug("Export index...")

        cursor = self.get_cursor()
        r = cursor.execute("SELECT * FROM solution").fetchall()

        if export_format == "JSON":
            json_dumps_list = []
            for solution in r:
                solution = dict(solution)
                self._append_metadata_to_solution_dict(solution)
                json_dumps_list.append(serialize_json(solution))

            write_dict_to_json(path, json_dumps_list)
        else:
            raise NotImplementedError('Unsupported format "%s"' % export_format)

        if close:
            self.close_current_connection()

    def __len__(self, close=True):
        cursor = self.get_cursor()
        r = cursor.execute("SELECT COUNT(*) FROM solution").fetchone()
        r = r[0]

        if close:
            self.close_current_connection()

        return r

    def _get_authors_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT a.* FROM author a "
            "JOIN solution_author sa ON sa.author_id = a.author_id "
            "WHERE sa.solution_id=:solution_id",
            {"solution_id": solution_id},
        ).fetchall()

        res = []
        for row in r:
            res.append(row["name"])

        if close:
            self.close_current_connection()

        return res

    def _get_tags_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT t.* FROM tag t "
            "JOIN solution_tag st ON st.tag_id = t.tag_id "
            "WHERE st.solution_id=:solution_id",
            {"solution_id": solution_id},
        ).fetchall()

        res = []
        for row in r:
            res.append(row["name"])

        if close:
            self.close_current_connection()

        return res

    def _get_citations_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT c.* FROM citation c "
            "JOIN solution_citation sc ON sc.citation_id = c.citation_id "
            "WHERE sc.solution_id=:solution_id",
            {"solution_id": solution_id},
        ).fetchall()

        res = []
        for row in r:
            res.append(dict(row))

        if close:
            self.close_current_connection()

        return res

    def _get_arguments_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT a.* FROM argument a "
            "JOIN solution_argument sa ON sa.argument_id = a.argument_id "
            "WHERE sa.solution_id=:solution_id",
            {"solution_id": solution_id},
        ).fetchall()

        res = []
        for row in r:
            row = dict(row)
            argument = {"name": row["name"], "type": row["type"]}
            if "description" in row and row["description"] is not None:
                argument["description"] = row["description"]
            if "required" in row and row["required"] is not None:
                argument["required"] = bool(row["required"])
            if "default_value" in row and row["default_value"] is not None:
                argument["default"] = row["default_value"]
            res.append(argument)

        if close:
            self.close_current_connection()

        return res

    def _get_custom_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT a.* FROM custom a "
            "JOIN solution_custom sa ON sa.custom_id = a.custom_id "
            "WHERE sa.solution_id=:solution_id",
            {"solution_id": solution_id},
        ).fetchall()

        res = {}
        for row in r:
            dict_row = dict(row)
            res[row["custom_key"]] = row["custom_value"]

        if close:
            self.close_current_connection()

        return res

    def _get_covers_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT c.* FROM cover c " "WHERE c.solution_id=:solution_id",
            {"solution_id": solution_id},
        ).fetchall()

        res = []
        for row in r:
            res.append(dict(row))

        if close:
            self.close_current_connection()

        return res

    def _get_documentation_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT d.* FROM documentation d " "WHERE d.solution_id=:solution_id",
            {"solution_id": solution_id},
        ).fetchall()

        res = []
        for row in r:
            res.append(row["documentation"])

        if close:
            self.close_current_connection()

        return res



import re
import tempfile
from pathlib import Path

import requests
from album.runner import album_logging
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from album.ci.utils.zenodo_api import ResponseStatus
from album.core.utils.operations.file_operations import (
    create_path_recursively,
    copy,
    check_zip,
)

module_logger = album_logging.get_active_logger


def is_downloadable(url):
    """Shows if url is a downloadable resource."""
    with _get_session() as s:
        h = s.head(url, allow_redirects=True)
        header = h.headers
        content_type = header.get("content-type")
        if "html" in content_type.lower():
            return False
        return True


def download_resource(url, path):
    """Downloads a resource given its url."""
    module_logger().debug("Download url %s to %s..." % (url, path))

    path = Path(path)

    if not is_downloadable(url):
        raise AssertionError('Resource "%s" not downloadable!' % url)

    r = _request_get(url)

    create_path_recursively(path.parent)
    with open(path, "wb") as f:
        for chunk in r:
            f.write(chunk)

    return path


def _get_session():
    s = requests.Session()
    retry = Retry(connect=3, backoff_factor=0.5)

    adapter = HTTPAdapter(max_retries=retry)

    s.mount("http://", adapter)
    s.mount("https://", adapter)

    return s


def _request_get(url):
    """Get a response from a request to a resource url."""
    with _get_session() as s:
        r = s.get(url, allow_redirects=True, stream=True)

        if r.status_code != ResponseStatus.OK.value:
            raise ConnectionError("Could not connect to resource %s!" % url)

        return r


def retrieve_redirect_url(url):
    with _get_session() as s:
        r = s.get(url, allow_redirects=True, stream=False)

        if r.status_code != ResponseStatus.OK.value:
            raise ConnectionError("Could not connect to resource %s!" % url)

        return r.url


def is_url(str_input: str):
    """Parses a url."""
    url_regex = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )
    return re.match(url_regex, str_input) is not None


def download(str_input, base):
    """Downloads a solution file into a temporary file."""
    Path(base).mkdir(exist_ok=True, parents=True)

    with _get_session() as s:
        r = s.get(str_input, allow_redirects=True, stream=True)

        new_file, tmp_file_name = tempfile.mkstemp(dir=base)
        with open(tmp_file_name, "wb") as out:
            out.write(r.content)
        if check_zip(tmp_file_name):
            new_file, tmp_file_name_zip = tempfile.mkstemp(dir=base, suffix=".zip")
            copy(tmp_file_name, tmp_file_name_zip)
            return Path(tmp_file_name_zip)
        return Path(tmp_file_name)



import io
import re
import subprocess
import sys
from typing import Optional

from album.runner import album_logging
from album.runner.album_logging import LogEntry, LogLevel

module_logger = album_logging.get_active_logger


class LogfileBuffer(io.StringIO):
    """Class for logging in a subprocess. Logs to the current active logger."""

    def __init__(
        self, module_logger, message_formatter="%(message)s", error_logger=False
    ):
        super().__init__()
        self.message_formatter = message_formatter
        self.leftover_message = None
        self.last_log = None
        self.is_error_logger = error_logger
        self.logger = module_logger
        self.logger_name = self.logger.name
        self.logger_name_script = self.logger_name + ".script"
        self.logger_name_current = self.logger_name_script
        self.logger_name_unnamed = self.logger_name_script + ".log"

    def write(self, input) -> int:
        if not isinstance(input, str):
            input = str(input, "utf-8")
        m = input.rstrip()

        log_entry = self.parse_log(m)

        if log_entry:

            if log_entry.name:
                log_entry.name = log_entry.name.strip()
                if log_entry.name.startswith("root.script"):
                    log_entry.name = log_entry.name.lstrip("root.script")
                    log_entry.name = log_entry.name.lstrip(".")
                if len(log_entry.name) > 0:
                    log_entry.name = self.logger_name_script + "." + log_entry.name
                else:
                    log_entry.name = self.logger_name_script
            else:
                log_entry.name = self.logger_name_unnamed

            if self.message_formatter and callable(self.message_formatter):
                message = self.message_formatter(log_entry.message)
            else:
                message = log_entry.message
            self.logger_name_current = log_entry.name
            self.logger.name = log_entry.name
            self._log(log_entry.level, message)
        else:  # unknown message not using or logging.
            self.logger.name = self.logger_name_current
            if self.is_error_logger:
                self._log("ERROR", m)
            else:
                self._log("INFO", m)
        self.logger.name = self.logger_name
        #
        return 1

    def _log(self, current_level, message):
        if LogLevel.INFO.name == current_level:
            self.logger.info(message)
        elif LogLevel.DEBUG.name == current_level:
            self.logger.debug(message)
        elif LogLevel.WARNING.name == current_level:
            self.logger.warning(message)
        elif LogLevel.ERROR.name == current_level:
            self.logger.error(message)
        else:
            self.logger.info(message)

    @staticmethod
    def parse_log(text) -> Optional[LogEntry]:
        res = LogfileBuffer._parse_album_runner_log(text)
        if not res:
            res = LogfileBuffer._parse_album_log(text)
            if not res:
                res = LogfileBuffer._parse_level_colon_log(text)
                if not res:
                    res = LogfileBuffer._parse_level_brackets_log(text)
        if res:
            if res.message:
                res.message = res.message.rstrip(" ")
            return res
        else:
            return None

    @staticmethod
    def get_script_logging_formatter_regex():
        regex_log_level = "DEBUG|INFO|WARNING|ERROR"
        return r"(%s)\s+([\s\S]+) - ([\s\S]+)?" % regex_log_level

    @staticmethod
    def _parse_album_runner_log(text):
        r = re.search(LogfileBuffer.get_script_logging_formatter_regex(), text)
        if r:
            return LogEntry(name=r.group(2), level=r.group(1), message=r.group(3))
        return None

    @staticmethod
    def _parse_album_log(text):
        r = re.search(
            r"\d\d:\d\d:\d\d (%s)(?:[\s]+(~*))? ([\s\S]+)?"
            % LogfileBuffer._regex_log_level(),
            text,
        )
        if r:
            if len(r.groups()) == 3:
                name = r.group(2)
                if name != None and len(name) == 0:
                    name = None
                return LogEntry(name=name, level=r.group(1), message=r.group(3))
            else:
                return LogEntry(name=None, level=r.group(1), message=r.group(2))
        return None

    @staticmethod
    def _parse_level_colon_log(text):
        r = re.search(r"(%s): ([\s\S]+)?" % LogfileBuffer._regex_log_level(), text)
        if r:
            return LogEntry(name=None, level=r.group(1), message=r.group(2))
        return None

    @staticmethod
    def _parse_level_brackets_log(text):
        r = re.search(r"\[(%s)\] ([\s\S]+)?" % LogfileBuffer._regex_log_level(), text)
        if r:
            return LogEntry(name=None, level=r.group(1), message=r.group(2))
        return None

    @staticmethod
    def _regex_log_level():
        return "DEBUG|INFO|WARNING|ERROR"


class LogProcessing:
    def __init__(self, logger, log_output, message_formatter):
        if log_output:
            self.info_logger = LogfileBuffer(logger, message_formatter)
            self.error_logger = LogfileBuffer(
                logger, message_formatter, error_logger=True
            )
        else:
            self.info_logger = io.StringIO()
            self.error_logger = io.StringIO()

    def log_info(self, s):
        self.info_logger.write(s)

    def log_error(self, s):
        self.error_logger.write(s)

    def close(self):
        self.info_logger.close()
        self.error_logger.close()


def run(command, log_output=True, message_formatter=None, pipe_output=True):
    """Runs a command in a subprocess thereby logging its output.

    Args:
        log_output:
            Indicates whether to log the output of the subprocess or not.
        message_formatter:
            Possibility to parse a lambda to format the message in a certain way.
        command:
            The command to run.
        pipe_output:
            Indicates whether to pipe the output of the subprocess or just return it as is.

    Returns:
        Exit status of the subprocess.

    Raises:
        RuntimeError:
            When exit-status of subprocess is not 0.

    """
    module_logger().debug("Running command: %s..." % " ".join(command))

    logger = album_logging.get_active_logger()
    log_processing = LogProcessing(logger, log_output, message_formatter)

    exit_status = _run_process(command, log_processing, pipe_output)

    return exit_status


class SubProcessError(RuntimeError):
    def __init__(self, exit_status, message) -> None:
        self.exit_status = exit_status
        super().__init__(message)


def _run_process(command, log: LogProcessing, pipe_output):
    if pipe_output:
        process = subprocess.Popen(
            command,
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )
        while True:
            output = process.stdout.readline()
            if process.poll() is not None:
                break
            if output:
                log.log_info(output.rstrip())
        sys.stdout.flush()
        rc = process.poll()
        if rc != 0:
            raise SubProcessError(rc, "ERROR while running %s" % " ".join(command))
        return rc
    else:
        return subprocess.run(command)


def check_output(command):
    """Runs a command thereby checking its output."""
    return subprocess.check_output(command).decode("utf-8")



import re
from typing import Dict
from typing import Optional


def is_release(line: str) -> bool:
    return line.startswith("## ")


def add_release(changes: Dict[str, dict], line: str) -> dict:
    release_line = line[3:].lower().strip(" ")
    # A release is separated by a space between version and release date
    # Release pattern should match lines like: "[0.0.1] - 2020-12-31" or [Unreleased]
    version, release_date = (
        release_line.split(" ", maxsplit=1)
        if " " in release_line
        else (release_line, None)
    )
    version = unlink(version)

    metadata = {"version": version, "release_date": extract_date(release_date)}
    try:
        metadata["semantic_version"] = to_semantic(version)
    except InvalidSemanticVersion:
        pass

    return changes.setdefault(version, {"metadata": metadata})


def unlink(value: str) -> str:
    return value.lstrip("[").rstrip("]")


def extract_date(date: str) -> str:
    if not date:
        return date

    return date.lstrip(" -(").rstrip(" )")


# Link pattern should match lines like: "[1.2.3]: https://github.com/user/project/releases/tag/v0.0.1"
link_pattern = re.compile(r"^\[(.*)\]: (.*)$")


def is_link(line: str) -> bool:
    return link_pattern.fullmatch(line) is not None


def to_raw_dict(changelog_path: str) -> Dict[str, dict]:
    changes = {}
    # As URLs can be defined before actual usage, maintain a separate dict
    urls = {}
    with open(changelog_path) as change_log:
        current_release = {}
        for line in change_log:
            clean_line = line.strip(" \n")

            if is_release(clean_line):
                current_release = add_release(changes, clean_line)
            elif is_link(clean_line):
                link_match = link_pattern.fullmatch(clean_line)
                urls[link_match.group(1).lower()] = link_match.group(2)
            elif clean_line:
                current_release["raw"] = current_release.get("raw", "") + line

    # Add url for each version (create version if not existing)
    for version, url in urls.items():
        changes.setdefault(version, {"metadata": {"version": version}})["metadata"][
            "url"
        ] = url

    unreleased_version = None
    for version, current_release in changes.items():
        metadata = current_release["metadata"]
        # If there is an empty release date, it identify the unreleased section
        if ("release_date" in metadata) and not metadata["release_date"]:
            unreleased_version = version

    changes.pop(unreleased_version, None)

    return changes


initial_semantic_version = {
    "major": 0,
    "minor": 0,
    "patch": 0,
    "prerelease": None,
    "buildmetadata": None,
}


class InvalidSemanticVersion(Exception):
    def __init__(self, version: str):
        super().__init__(
            f"{version} is not following semantic versioning. Check https://semver.org for more information."
        )


semantic_versioning = re.compile(
    r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:[-\.]?(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)


def to_semantic(version: Optional[str]) -> dict:
    if not version:
        return initial_semantic_version.copy()

    match = semantic_versioning.fullmatch(version)
    if match:
        return {
            key: int(value) if key in ("major", "minor", "patch") else value
            for key, value in match.groupdict().items()
        }

    raise InvalidSemanticVersion(version)



from abc import ABCMeta, abstractmethod


class ICloneManager:
    """Interface handling the creation of new catalogs and solutions."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def clone(
        self,
        path: str,
        target_dir: str,
        name: str,
        git_email: str = None,
        git_name: str = None,
    ) -> None:
        """
        Function corresponding to the `clone` subcommand of `album`.

        Args:
            path: the source of the clone command - a solution (group:name:version, path, or URL to file) or a catalog
                template string (i.e. template:catalog)
            target_dir: the directory where the cloned solution or catalog will be added to
            name: the name of the solution or catalog to be created
            git_email: (catalog cloning only) email of the git user to be used for pushing to target_dir (in case it's already a git repo)
            git_name: (catalog cloning only) name of the git user to be used for pushing to target_dir (in case it's already a git repo)

        """
        raise NotImplementedError



from abc import ABCMeta, abstractmethod


class ITestManager:
    """Interface managing testing routine of a solution. Similar to the installation process, a configured \"test \"
    routine of a solution is executed in the target environment (The conda environment the solution lives in).
    Solutions must be installed to run their testing routine."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def test(self, solution_to_resolve: str, args=None):
        """Function corresponding to the `test` subcommand of `album`."""
        raise NotImplementedError



import os

from album.core.utils.operations.resolve_operations import dict_to_coordinates
from album.runner import album_logging

from album.ci.utils import zenodo_api

module_logger = album_logging.get_active_logger


class ZenodoManager:
    def __init__(self, zenodo_base_url, zenodo_access_token):
        self.query = zenodo_api.ZenodoAPI(zenodo_base_url, zenodo_access_token)

    @staticmethod
    def zenodo_upload(deposit, file):
        """Uploads a solution file to a ZenodoDeposit. Expects the deposit to be writable. (e.g. unpublished)

        Args:
            deposit:
                The deposit to upload to.
            file:
                The file to upload.

        Returns:
            The updated deposit.

        """
        file_basename = os.path.basename(file)

        if file_basename in deposit.files:  # File does exist
            module_logger().debug(
                "Update file %s in Zenodo deposit with id %s..."
                % (file_basename, deposit.id)
            )
            deposit.update_file(file_basename, file)
        else:
            module_logger().debug(
                "Create file %s in Zenodo deposit with id %s..."
                % (file_basename, deposit.id)
            )
            deposit.create_file(file)

        return deposit

    @staticmethod
    def zenodo_delete(deposit, file):
        """Deletes a solution file from a ZenodoDeposit. Expects the deposit to be writable. (e.g. unpublished)

        Args:
            deposit:
                The deposit to upload to.
            file:
                The file to delete.

        Returns:
            The updated deposit.

        """
        file_basename = os.path.basename(file)

        if file_basename in deposit.files:  # File does exist
            module_logger().debug(
                "Update file %s in Zenodo deposit with id %s..."
                % (file_basename, deposit.id)
            )
            deposit.delete_file(file_basename, file)
        else:
            module_logger().warning(
                "Cannot find file %s in Zenodo deposit with id %s."
                % (file_basename, deposit.id)
            )

        return deposit

    def zenodo_get_deposit(self, metadata, deposit_id, expected_files=None):
        """Queries zenodo to get the deposit of the solution_file. Creates an empty deposit if no deposit exists.

        Args:
            metadata:
                The metadata of the zenodo deposit.
            deposit_id:
                The deposit ID of the deposit the solution file lives in.
            expected_files:
                The files the deposit should hold.

        Returns:
            The deposit. Either brand-new, or the one holding the solution file.

        Raises:
            AttributeError:
               When the deposit does NOT hold the file it should hold.
               When the deposit_name given does not match the deposit name found

        """

        if deposit_id:  # case deposit already exists
            deposit = self._zenodo_get_deposit_by_id(deposit_id)

            self._check_deposit(deposit, metadata.title, expected_files)

        else:
            module_logger().debug("Query new deposit with DOI...")
            deposit = self.query.deposit_create_with_prereserve_doi(metadata)

        return deposit

    def _get_deposit_name(self, solution_meta):
        if solution_meta["title"]:
            deposit_name = solution_meta["title"]
        else:
            deposit_name = str(dict_to_coordinates(solution_meta))
        return deposit_name

    def zenodo_get_unpublished_deposit_by_id(
        self, deposit_id, deposit_name, expected_files=None
    ):
        deposit = self._zenodo_get_unpublished(deposit_id)

        # raise error when not found
        if not deposit:
            raise RuntimeError("Deposit with id %s not found!" % deposit_id)

        self._check_deposit(deposit, deposit_name, expected_files)

        return deposit

    @staticmethod
    def _check_deposit(deposit, expected_deposit_name, expected_files=None):
        # assess files in deposit.
        if expected_files:
            for expected_file in expected_files:
                if expected_file not in deposit.files:
                    raise AttributeError(
                        "Deposit has no file with the name %s!" % expected_file
                    )

        if deposit.title != expected_deposit_name:
            raise AttributeError(
                "Deposit name different than expected! Expected: %s Found: %s"
                % (expected_deposit_name, deposit.title)
            )

    def _zenodo_get_deposit_by_id(self, deposit_id):
        """Querying deposit by id."""
        # get published deposit
        deposit = self._zenodo_get_published(deposit_id)

        # get unpublished deposit
        if not deposit:
            deposit = self._zenodo_get_unpublished(deposit_id)

        # raise error when not found
        if not deposit:
            raise RuntimeError("Deposit with id %s not found!" % deposit_id)

        return deposit

    def _zenodo_get_unpublished(self, deposit_id):
        deposit = self.query.deposit_get(
            deposit_id, status=zenodo_api.DepositStatus.DRAFT
        )

        if deposit:
            deposit = deposit[0]

            module_logger().debug("Deposit with id %s found..." % deposit.id)

            if (
                deposit.metadata is None
                or not isinstance(deposit.metadata.prereserve_doi, dict)
                or "doi" not in deposit.metadata.prereserve_doi.keys()
            ):
                raise RuntimeError("Deposit has no prereserved DOI! Invalid deposit!")

        return deposit

    def _zenodo_get_published(self, deposit_id):
        deposit = self.query.deposit_get(deposit_id)

        if deposit and deposit[0].submitted:
            deposit = deposit[0]

            module_logger().debug(
                "Deposit with id %s found but it has already been published. Querying new version..."
                % deposit.id
            )

            deposit = deposit.new_version()

        return deposit
